/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.datatypes;

import ij.process.*;
import ij.*;
import java.io.*;
import java.util.*;

import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.operator.MTBData;

/**
 * Class for generating or extracting intensity histograms of ImageJ
 * ImageProcessor, ImageJ ImageStack and MTBImage
 * 
 * @modified Sep 29, 2010
 * 
 * @author glass
 */
public class MTBHistogram extends MTBData {
		/**
		 * Array holding the histogram data.
		 */
		private double[] data;
		/**
		 * Number of entries in the histogram array (number of bins).
		 */
		private int size;
		/**
		 * Smallest intensity value contained in the input image.
		 */
		private int smallestNonEmptyBin;
		/**
		 * Largest intensity value contained in the input image.
		 */
		private int largestNonEmptyBin;
		/**
		 * Sum of all histogram entries.
		 */
		private double sum;
		/**
		 * Lower boundary of the first bin.
		 */
		private double m_lowBound;
		/**
		 * Upper boundary of the last bin
		 */
		private double m_highBound;

		/**
		 * Construct a new MTBHistogram object from the given MTBImage.
		 * 
		 * @param img
		 *          MTBImage
		 */
		public MTBHistogram(MTBImage img) {
				m_lowBound = 0;
				m_highBound = 0;
				int sizeX = img.getSizeX();
				int sizeY = img.getSizeY();
				int sizeZ = img.getSizeZ();
				int sizeT = img.getSizeT();
				int sizeC = img.getSizeC();
				double[] h;
				// determine the appropriate histogram size
				switch (img.getType()) {
				case MTB_BYTE: // 8 Bit greyscale
						h = new double[256];
						break;
				case MTB_SHORT: // 16 Bit greyscale
						h = new double[65536];
						break;
				// TODO: add support for the other MTBImage-subtypes
				default:
						System.err.println("Histogram: Image Type " + img.getType()
						    + " not yet supported");
						// System.exit(-1);
						h = new double[256];
				}
				// collect the data
				for (int x = 0; x < sizeX; x++) {
						for (int y = 0; y < sizeY; y++) {
								for (int z = 0; z < sizeZ; z++) {
										for (int t = 0; t < sizeT; t++) {
												for (int c = 0; c < sizeC; c++) {
														h[(int) img.getValueDouble(x, y, z, t, c)]++;
												}
										}
								}
						}
				}
				initialize(h);
		}

		/**
		 * Construct a new MTBHistogram object from the given ImageJ ImageProcessor.
		 * 
		 * @param ip
		 *          ImageJ ImageProcessor
		 */
		public MTBHistogram(ImageProcessor ip) {
				m_lowBound = 0;
				m_highBound = 0;
				int[] h = ip.getHistogram();
				initialize(h);
		}

		/**
		 * Construct a new MTBHistogram object from the given ImageJ ImageStack.
		 * 
		 * @param stack
		 *          ImageJ ImageStack
		 */
		public MTBHistogram(ImageStack stack) {
				m_lowBound = 0;
				m_highBound = 0;
				ImageProcessor temp = stack.getProcessor(1);
				int[] t = temp.getHistogram();
				double[] h = new double[t.length];
				for (int i = 1; i <= stack.getSize(); i++) {
						temp = stack.getProcessor(i);
						t = temp.getHistogram();
						for (int j = 0; j < t.length; j++) {
								h[j] += t[j];
						}
				}
				initialize(h);
		}

		/**
		 * Construct a new MTBHistogram object from the given integer array.
		 * 
		 * @param h
		 *          histogram array containing integer values
		 */
		public MTBHistogram(int[] h) {
				m_lowBound = 0;
				m_highBound = 0;
				initialize(h);
		}

		/**
		 * Construct a new MTBHistogram object from the given double array.
		 * 
		 * @param h
		 *          histogram array containing double values
		 */
		public MTBHistogram(double[] h) {
				m_lowBound = 0;
				m_highBound = 0;
				initialize(h);
		}

		/**
		 * Construct a new MTBHistogram object. Extract histogram of MTBImage with
		 * specified number of bins.
		 * 
		 * @param img
		 *          MTBImage
		 * @param bins
		 *          number of bins
		 */
		public MTBHistogram(MTBImage img, int bins) {
				m_lowBound = 0;
				m_highBound = 0;
				int sizeX = img.getSizeX();
				int sizeY = img.getSizeY();
				int sizeZ = img.getSizeZ();
				int sizeT = img.getSizeT();
				int sizeC = img.getSizeC();
				double[] h;
				// determine the appropriate histogram size
				switch (img.getType()) {
				case MTB_BYTE: // 8 Bit greyscale
						h = new double[256];
						break;
				case MTB_DOUBLE: // 16 Bit greyscale
						h = new double[65536];
						break;
				// TODO: add support for the other MTBImage-subtypes
				default:
						System.err.println("Histogram: Image Type " + img.getType()
						    + " not yet supported");
						// System.exit(-1);
						h = new double[256];
				}
				// collect the data
				for (int x = 0; x < sizeX; x++) {
						for (int y = 0; y < sizeY; y++) {
								for (int z = 0; z < sizeZ; z++) {
										for (int t = 0; t < sizeT; t++) {
												for (int c = 0; c < sizeC; c++) {
														h[(int) img.getValueDouble(x, y, z, t, c)]++;
												}
										}
								}
						}
				}
				initialize(h);
				binning(bins);
		}

		/**
		 * Constructor for all MTBImage types except RGB type with specification of
		 * binning and value range.
		 * 
		 * @param img
		 *          input image for which the histogram is calculated
		 * @param bins
		 *          number of histogram bins
		 * @param lowBoundary
		 *          determines the lower boundary of the first bin
		 * @param highBoundary
		 *          determines the upper boundary of the last bin
		 * @throws IllegalArgumentException
		 *           if an RGB image is given as input
		 */
		public MTBHistogram(MTBImage img, int bins, double lowBoundary,
		    double highBoundary) throws IllegalArgumentException {
				if (img.getType() == MTBImageType.MTB_RGB) {
						throw new IllegalArgumentException(
						    "Histogram creation for RGB images not supported.");
				}
				if (lowBoundary >= highBoundary) {
						throw new IllegalArgumentException(
						    "Histogram  must be larger than lowest value");
				}
				m_lowBound = lowBoundary;
				m_highBound = highBoundary;
				data = new double[bins];
				int sizeX = img.getSizeX();
				int sizeY = img.getSizeY();
				int sizeStack = img.getSizeStack();
				double[] h = new double[bins];
				// get the actual slice index of the image
				int sliceIdx = img.getActualSliceIndex();
				// collect the data
				for (int i = 0; i < sizeStack; i++) {
						img.setActualSliceIndex(i);
						for (int y = 0; y < sizeY; y++) {
								for (int x = 0; x < sizeX; x++) {
										h[this.getBinIndex(img.getValueDouble(x, y))] += 1.0;
								}
						}
				}
				// restore the slice index of the image
				img.setActualSliceIndex(sliceIdx);
				this.initialize(h);
		}

		/**
		 * Determines the index of the histogram bin, to which a value is assigned, if
		 * histogram value range and binning were specified at creation.
		 * 
		 * @param value
		 *          image value
		 * @return index of histogram bin for a given value (indices range from 0 to
		 *         (bins-1))
		 */
		public int getBinIndex(double value) {
				if (m_lowBound == m_highBound) {
						return (int) Math.round(value);
				} else {
						int idx = (int) ((value - m_lowBound) * (double) data.length / (m_highBound - m_lowBound));
						if (idx >= data.length) {
								idx = data.length - 1;
						} else if (idx < 0) {
								idx = 0;
						}
						return idx;
				}
		}

		/**
		 * Determines the value midpoint of a histogram bin for given bin index, if
		 * histogram value range and binning were specified at creation.
		 * 
		 * @param binIndex
		 *          bin index
		 * @return (gray value) midpoint of a bin
		 */
		public double getBinMidpoint(int binIndex) {
				if (m_lowBound == m_highBound) {
						return binIndex;
				} else {
						return (m_lowBound + (m_highBound - m_lowBound) / (double) data.length
						    * ((double) binIndex + 0.5));
				}
		}

		@Override
		public String toString() {
				String s = "";

				for (int i = 0; i < data.length; i++) {
						s += (this.getBinMidpoint(i) + " " + data[i] + "\n");
				}
				return s;
		}

		/**
		 * Add histogram data to this MTBHistogram object.
		 * 
		 * @param h
		 *          MTBHistogram to add
		 * @throws IllegalArgumentException
		 */
		public void add(MTBHistogram h) throws IllegalArgumentException {
				if (this.data.length != h.data.length || this.m_lowBound != h.m_lowBound
				    || this.m_highBound != h.m_highBound) {
						throw new IllegalArgumentException(
						    "Input histogram doesn't match the histogram it is added to.");
				}
				this.sum += h.sum;
				for (int i = 0; i < size; i++) {
						this.data[i] += h.data[i];
				}
		}

		/**
		 * Construct a new MTBHistogram object. Extract histogram of ImageJ
		 * ImageProcessor with specified number of bins.
		 * 
		 * @param ip
		 *          ImageJ ImageProcessor
		 * @param bins
		 *          number of bins
		 */
		public MTBHistogram(ImageProcessor ip, int bins) {
				m_lowBound = 0;
				m_highBound = 0;
				int[] h = ip.getHistogram();
				initialize(h);
				binning(bins);
		}

		/**
		 * Construct a new MTBHistogram object. Extract histogram of ImageJ ImageStack
		 * with specified number of bins.
		 * 
		 * @param stack
		 *          ImageJ ImageStack
		 * @param bins
		 *          number of bins
		 */
		public MTBHistogram(ImageStack stack, int bins) {
				m_lowBound = 0;
				m_highBound = 0;
				ImageProcessor temp = stack.getProcessor(1);
				int[] t = temp.getHistogram();
				double[] h = new double[t.length];
				for (int i = 1; i <= stack.getSize(); i++) {
						temp = stack.getProcessor(i);
						t = temp.getHistogram();
						for (int j = 0; j < t.length; j++) {
								h[j] += t[j];
						}
				}
				initialize(h);
				binning(bins);
		}

		/**
		 * Construct a new MTBHistogram object. Create histogram from array with
		 * specified number of bins.
		 * 
		 * @param h
		 *          histogram array
		 * @param bins
		 *          number of bins
		 */
		public MTBHistogram(int[] h, int bins) {
				m_lowBound = 0;
				m_highBound = 0;
				initialize(h);
				binning(bins);
		}

		/**
		 * Construct a new MTBHistogram object. Create histogram from array with
		 * specified number of bins.
		 * 
		 * @param h
		 *          histogram array
		 * @param bins
		 *          number of bins
		 */
		public MTBHistogram(double[] h, int bins) {
				m_lowBound = 0;
				m_highBound = 0;
				initialize(h);
				binning(bins);
		}

		/**
		 * Construct a new MTBHistogram object. Read histogram from text file.
		 * 
		 * @param fileName
		 *          file containing the histogram data
		 */
		public MTBHistogram(String fileName) {
				m_lowBound = 0;
				m_highBound = 0;
				try {
						File f = new File(fileName);
						FileReader in = new FileReader(f);
						BufferedReader br = new BufferedReader(in);
						Vector<Double> data = new Vector<Double>();
						String line = br.readLine();
						StringTokenizer strtok = new StringTokenizer(line, " ");
						while (strtok.hasMoreTokens()) {
								data.addElement(Double.parseDouble(strtok.nextToken()));
						}
						br.close();
						int size = data.size();
						double[] h = new double[size];
						for (int i = 0; i < size; i++) {
								h[i] = data.get(i).doubleValue();
						}
						initialize(h);
				} catch (IOException e) {
						System.err.println(e);
				}
		}

		/**
		 * Initialize histogram, i.e. convert values to double, determine smallest non
		 * empty bin, largest non empty bin and sum of the histogram.
		 * 
		 * @param h
		 *          initializing values in a integer array
		 */
		private void initialize(int[] h) {
				size = h.length;
				data = new double[size];
				// search for first non-empty entry
				for (int i = 0; i < size; i++) {
						if (h[i] != 0) {
								smallestNonEmptyBin = i;
								break;
						}
				}
				// search for last non-empty entry
				for (int i = size - 1; i >= 0; i--) {
						if (h[i] != 0) {
								largestNonEmptyBin = i;
								break;
						}
				}
				// calculate the sum of all entries
				sum = 0;
				for (int i = smallestNonEmptyBin; i <= largestNonEmptyBin; i++) {
						data[i] = h[i];
						sum += h[i];
				}
		}

		/**
		 * Initialize histogram, i.e. determine smallest non empty bin, largest non
		 * empty bin and sum of the histogram.
		 * 
		 * @param h
		 *          initializing values in a double array
		 */
		private void initialize(double[] h) {
				size = h.length;
				data = new double[size];
				// search for first non-empty entry
				for (int i = 0; i < size; i++) {
						if (h[i] != 0) {
								smallestNonEmptyBin = i;
								break;
						}
				}
				// search for last non-empty entry
				for (int i = size - 1; i >= 0; i--) {
						if (h[i] != 0) {
								largestNonEmptyBin = i;
								break;
						}
				}
				// calculate the sum of all entries
				sum = 0;
				for (int i = smallestNonEmptyBin; i <= largestNonEmptyBin; i++) {
						data[i] = h[i];
						sum += h[i];
				}
		}

		/**
		 * Normalize the histogram, so that the contoining values sums up to 1.
		 */
		public void normalize() {
				for (int i = 0; i < size; i++) {
						data[i] /= sum;
				}
				sum = 1;
		}

		/**
		 * Normalize the histogram without changing the sum of absolute frequencies
		 * (normalizing factor), because normalize() calls initialize() and thus
		 * destroys this normalizing factor.
		 */
		public void normalizeOnly() {
				for (int i = 0; i < size; i++) {
						data[i] /= sum;
				}
		}

		/**
		 * Truncate the histogram to the specified interval.
		 * 
		 * @param l
		 *          left border
		 * @param r
		 *          right border
		 */
		public void truncate(int l, int r) {
				double[] temp = new double[r - l + 1];
				for (int i = l; i <= r; i++) {
						temp[i - l] = data[i];
				}
				initialize(temp);
		}

		/**
		 * Change the binning of the histogram.
		 * 
		 * @param bins
		 *          number of bins
		 */
		public void binning(int bins) {
				double[] temp = new double[bins];
				if (size % bins == 0) // binsize integer
				{
						// System.out.println("integer");
						int binsize = size / bins;
						if (binsize > 1) {
								for (int i = 0; i < size; i++) {
										double value = data[i];
										temp[i / binsize] += value;
								}
						} else // number of bins larger than size
						{
								// first size bins keep their values, following bins get 0
								for (int i = 0; i < size; i++) {
										double value = data[i];
										temp[i] += value;
								}
						}
				} else // binsize not integer
				{
						// System.out.println("not integer");
						double binsize = (double) size / bins;
						double[] borders = new double[bins];
						if (binsize >= 1) {
								for (int i = 0; i < bins; i++) {
										borders[i] = (binsize * (i + 1)) - 1; // -1 because [0, ..., size - 1]
										System.out.println(borders[i]);
								}
								int x = 0;
								for (int i = 0; i < size; i++) {
										int bin = (int) (i / binsize);
										double value = data[i];
										temp[bin] += value;
										double f = borders[x] - i;
										// System.out.println("f: " + f);
										if (f < 1) {
												x++;
												if (f > 0) {
														// System.out.println("i " + i + " bin: " + bin);
														temp[bin] += f * data[i + 1];
														temp[bin + 1] += (1 - f) * data[i + 1];
														i++;
												}
										}
								}
						} else // number of bins larger than size
						{
								// first size bins keep their values, following bins get 0
								for (int i = 0; i < size; i++) {
										double value = data[i];
										temp[i] += value;
								}
						}
				}
				initialize(temp);
		}

		/**
		 * Calculate the logarithmic histogram zero entries remain zero!
		 */
		public void logarithmize() {
				double[] temp = new double[size];
				for (int i = 0; i < size; i++) {
						double value = data[i];
						// log([value <= 1]) here remains 0
						if (value > 1) {
								temp[i] = Math.log(value);
						}
				}
				initialize(temp);
		}

		/**
		 * Calculate the cumulative histogram.
		 */
		public void cumulate() {
				double[] temp = new double[size];
				double csum = 0;
				for (int i = 0; i < size; i++) {
						double value = data[i];
						csum += value;
						temp[i] = csum;
				}
				initialize(temp);
		}

		/**
		 * Calculate the cumulative histogram without changing the sum of absolute
		 * frequencies (normalizing factor), because cumulate() calls initialize() and
		 * thus destroys this normalizing factor
		 */
		public void cumulateOnly() {
				double sum = 0.0;
				for (int i = 0; i < size; i++) {
						sum += data[i];
						data[i] = sum;
				}
		}

		/**
		 * Return the histogram data values.
		 * 
		 * @return Data values.
		 */
		public double[] getData() {
				return data;
		}

		/**
		 * Returns the number of bins in the histogram.
		 * 
		 * @return Number of bins.
		 */
		public int getSize() {
				return size;
		}

		/**
		 * Returns the smallest non empty bin.
		 * 
		 * @return Smallest non empty bin.
		 */
		public int getSmallestNonEmptyBin() {
				return smallestNonEmptyBin;
		}

		/**
		 * Returns the largest non empty bin.
		 * 
		 * @return Largest non empty bin.
		 */
		public int getLargestNonEmptyBin() {
				return largestNonEmptyBin;
		}

		/**
		 * Returns global contrast, i.e. subtraction of smallest non empty bin from
		 * the largest non empty bin.
		 * 
		 * @return Global contrast.
		 */
		public int getGlobalContrast() {
				return (largestNonEmptyBin - smallestNonEmptyBin) + 1;
		}

		/**
		 * Returns the sum of all histogram entries.
		 * 
		 * @return Sum of entries.
		 */
		public double getSum() {
				return sum;
		}

		/**
		 * Returns the bin value at i-th position.
		 * 
		 * @param i
		 *          position
		 * @return Bin value.
		 */
		public double getBinValue(int i) {
				if (i >= 0 && i < size) {
						return data[i];
				} else {
						return 0;
				}
		}

		/**
		 * Set the bin value at i-th position.
		 * 
		 * @param i
		 *          position
		 * @param val
		 *          bin value
		 */
		public void setBinValue(int i, double val) {
				if (i >= 0 && i < size) {
						data[i] = val;
				}
				initialize(data);
		}

		/**
		 * Returns the largest value in the histogram.
		 * 
		 * @return Largest value.
		 */
		public double getMaxValue() {
				double maxValue = 0;
				for (int i = 0; i < size; i++) {
						if (data[i] > maxValue) {
								maxValue = data[i];
						}
				}
				return maxValue;
		}

		/**
		 * Returns the bin containing the largest value in the histogram,
		 * 
		 * @return Bin with largest value.
		 */
		public double getMaxValueBin() {
				double maxValue = 0;
				int maxValueBin = 0;
				for (int i = 0; i < size; i++) {
						if (data[i] > maxValue) {
								maxValue = data[i];
								maxValueBin = i;
						}
				}
				return maxValueBin;
		}

		/**
		 * Return the mean of all intensity values.
		 * 
		 * @return Mean of intensity.
		 */
		public double getMean() {
				double mean = 0;
				for (int i = 0; i < size; i++) {
						mean += data[i] * i;
				}
				return (mean / sum);
		}

		/**
		 * Return the variance of all intensity values.
		 * 
		 * @return Variance of intensity.
		 */
		public double getVariance() {
				double var = 0;
				double m = getMean();
				for (int i = 0; i < size; i++) {
						var += ((i - m) * (i - m)) * data[i];
				}
				return (var / sum);
		}

		/**
		 * Write the histogram into a text file in a format that can be read and
		 * further processed by Matlab/ GNU Octave for example.
		 * 
		 * @param fileName
		 *          file name for histogram data
		 */
		public void save(String fileName) {
				// write histogram into file
				try {
						BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
						for (int i = 0; i < data.length; i++) {
								bw.write(data[i] + " ");
						}
						bw.newLine();
						bw.close();
				} catch (FileNotFoundException e) {
						System.err.println(e);
				} catch (IOException e) {
						System.err.println(e);
				}
		}

		// /**
		// * scales values of the histogram data to [1,..., K], logarithmizes them
		// * and scales them back to the initial range
		// *
		// * @param K upper bound of histogram values
		// */
		// public void scaleLog(int K)
		// {
		// double max = getMaxValue();
		//		
		// for(int i = 0; i < size; i++)
		// {
		// double scaleVal = 1 + (((K-1) / (max)) * data[i]);
		// double logVal = (max/Math.log(K)) * Math.log(scaleVal);
		// data[i] = logVal;
		// }
		// }

		/**
		 * Create histogram from one-dimensional array containing intensity values.
		 * 
		 * @param pixels
		 *          one-dimensional pixel array
		 * @param size
		 *          size of the new histogram
		 */
		public static MTBHistogram createHistogramFromPixelArray(int[] pixels,
		    int size) {
				double[] data = new double[size];
				for (int i = 0; i < pixels.length; i++) {
						int pos = pixels[i];
						data[pos]++;
				}
				return new MTBHistogram(data);
		}
}
