/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.datatypes;

import java.util.Vector;

import de.unihalle.informatik.MiToBo.operator.MTBData;


/** A set of 3D regions living in a common domain which is
  * a rectangular subset of R x R.
  * Each regions is of type Region3D 
  *
  * @author Stefan Posch
 */

public class  MTBRegion3DSet extends MTBData {

	private boolean debug = false;

	// subset of R x R	
	/** Minimal x coordinate of the domain of this region set.
	 */
	private double xMin;

	/** Minimal y coordinate of the domain of this region set.
	 */
	private double yMin;
	
	/** Minimal z coordinate of the domain of this region set.
	 */
	private double zMin;
	
	/** Maximal x coordinate of the domain of this region set.
	 */
	private double xMax;

	/** Maximal y coordinate of the domain of this region set.
	 */
	private double yMax;

	/** Maximal y coordinate of the domain of this region set.
	 */
	private double zMax;
	
	/** The set of regions represented as a Vector.
	 */
	Vector<MTBRegion3D>	regionSet;

	/** Construct an empty set of regions with given extent of domain
	 */
	public MTBRegion3DSet( double xMin, double yMin, double zMin, double xMax, double yMax, double zMax) {
		this.xMin = xMin;
		this.yMin = yMin;
		this.zMin = zMin;
		this.xMax = xMax;
		this.yMax = yMax;
		this.zMax = zMax;
		

		regionSet = new Vector<MTBRegion3D> ();

		setProperty( "xMin", xMin);
		setProperty( "yMin", yMin);
		setProperty( "zMin", zMin);
		setProperty( "xMax", xMax);
		setProperty( "yMax", yMax);
		setProperty( "zMax", zMax);
	}
	
	/** Construct an empty set (with memory allocated for N elements) of regions with given extent of domain
	 */
	public MTBRegion3DSet(int N, double xMin, double yMin, double zMin, double xMax, double yMax, double zMax) {
		this.xMin = xMin;
		this.yMin = yMin;
		this.zMin = zMin;
		this.xMax = xMax;
		this.yMax = yMax;
		this.zMax = zMax;
		

		regionSet = new Vector<MTBRegion3D>(N);

		setProperty( "xMin", xMin);
		setProperty( "yMin", yMin);
		setProperty( "zMin", zMin);
		setProperty( "xMax", xMax);
		setProperty( "yMax", yMax);
		setProperty( "zMax", zMax);
	}
	
	/** Construct an empty set of regions with given extent of domain
	 */
	public MTBRegion3DSet( Vector<MTBRegion3D> regions, double xMin, double yMin, double zMin, double xMax, double yMax, double zMax) {
		this.xMin = xMin;
		this.yMin = yMin;
		this.zMin = zMin;
		this.xMax = xMax;
		this.yMax = yMax;
		this.zMax = zMax;

		regionSet = new Vector<MTBRegion3D> ();
		for ( int i=0; i < regions.size() ; i++ ) 
			add( regions.elementAt(i));

		setProperty( "xMin", xMin);
		setProperty( "yMin", yMin);
		setProperty( "zMin", zMin);
		setProperty( "xMax", xMax);
		setProperty( "yMax", yMax);
		setProperty( "zMax", zMax);
	}

	public double getXmin() {
		return this.xMin;
	}
	public double getXmax() {
		return this.xMax;
	}
	public double getYmin() {
		return this.yMin;
	}
	public double getYmax() {
		return this.yMax;
	}
	public double getZmin() {
		return this.zMin;
	}
	public double getZmax() {
		return this.zMax;
	}
	
	/** Get the number of regions of this region set
	 *  @return number of regions
	 */
	public int size() {
		return regionSet.size();
	}

	/**
	 * Remove all regions from this set
	 */
	public void clear() {
		this.regionSet.clear();
	}
	
	/** Get a region by index
	 *
	 * @return i-th region
	 */
	public MTBRegion3D get( int i) {
		return this.regionSet.get(i);
	}
	
	/** Get a region by index
	 *
	 * @return i-th region
	 */
	public MTBRegion3D elementAt( int i) {
		if ( i < 0 || i >= size() )
			throw new ArrayIndexOutOfBoundsException();

		return regionSet.elementAt( i);
	}

	/** Append a region (at the end) to the set of regions.
	 * @param 	region	region to add
	 */
	public boolean add( MTBRegion3D region) {
		return regionSet.add( region);
	}

	/** Set the i-th region from the set
	 * @param 	i	index of region to remove
	 */
	public void setElementAt(MTBRegion3D region, int i) {
		regionSet.setElementAt(region, i);
	}

	/** Delete the i-th region from the set
	 * @param 	i	index of region to remove
	 */
	public void removeElementAt(int i) {
		regionSet.removeElementAt(i);
	}

}
