/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.io.files;

import it.geosolutions.imageio.plugins.tiff.TIFFImageReadParam;
import it.geosolutions.imageioimpl.plugins.tiff.TIFFImageReader;

import java.awt.image.DataBuffer;
import java.awt.image.RenderedImage;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.stream.ImageInputStream;

import org.graphdrawing.graphml.xmlns.DataType;
import org.graphdrawing.graphml.xmlns.GraphType;
import org.graphdrawing.graphml.xmlns.GraphmlType;
import org.graphdrawing.graphml.xmlns.NodeType;
import org.graphdrawing.graphml.xmlns.PropertyListType;
import org.graphdrawing.graphml.xmlns.PropertyType;

import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageRGB;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * Operator for reading MTBImages from disk.
 * The image must be in TIFF format.
 * The history graph is read automatically from the file of the same file basename and .mph extension
 * @author Oliver Gress
 *
 */
public class ReaderMTBImage extends MTBOperator {

	/**
	 * Constructor.
	 * @param filename
	 * @throws MTBOperatorException
	 */
	public ReaderMTBImage(String filename) throws MTBOperatorException {

		this.setFileName(filename);
		
		this.collectArguments();
	}

	@Override
	protected void collectArguments() throws MTBOperatorException {
		super.collectArguments();


		MTBOpArgumentDescriptor[] inputDescriptors = null;

		
		MTBOpArgumentDescriptor[] outputDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
			    "resultImage", MTBImage.class, "Image that has been read from disk", true, null) };

		
		MTBOpArgumentDescriptor[] parameterDescriptors = new MTBOpArgumentDescriptor[] {
		    new MTBOpArgumentDescriptor("fileName", java.lang.String.class,
		        "Image file name", true, null)/*, 
		    new MTBOpArgumentDescriptor("sliceList", java.util.Vector.class,
			    "(Optional) List of slices to be loaded", false, null)*/	    
		};
		

		MTBOpArgumentDescriptor[] supplementalDescriptor = null;

		
		this.addDescriptors(parameterDescriptors, inputDescriptors, outputDescriptors,
		    supplementalDescriptor);
	}
	
	/**
	 * Set resulting image
	 * @param img
	 * @throws MTBOperatorException
	 */
	protected void setResultImage(MTBImage img) throws MTBOperatorException {
		this.setOutput("resultImage", img);
	}
	
	/**
	 * Get resulting image that was read from disk.
	 * @return
	 * @throws MTBOperatorException
	 */
	public MTBImage getResultImage() throws MTBOperatorException {
		return (MTBImage)this.getOutput("resultImage");
	}
	
	/**
	 * Set the filename of the image that has to be read.
	 * @param filename
	 * @throws MTBOperatorException
	 */
	public void setFileName(String filename) throws MTBOperatorException {
		this.setParameter("fileName", filename);
	}
	
	/**
	 * Get the filename of the image that has to be read.
	 * @return
	 * @throws MTBOperatorException
	 */
	public String getFileName() throws MTBOperatorException {
		return (String)this.getParameter("fileName");
	}
	
//	/**
//	 * Set a list of (z,t,c) coordinates to specify the slices to be loaded.
//	 * Not the whole image data is read, but only the specified slices. Because the
//	 * structure of a hyperstack cannot be preserved, all slices are stored in a 
//	 * z-stack, thus loosing hyperstack information. The user must himself reconstruct
//	 * hyperstack-structure if needed.
//	 * Set null (default) to read the whole image with all dimension information.
//	 * @param slicelist
//	 * @throws MTBOperatorException
//	 */
//	public void setSliceList(Vector<MTBPoint3D> slicelist) throws MTBOperatorException {
//		this.setParameter("sliceList", slicelist);
//	}
//	
//	/**
//	 * Get the list of slice coordinates (see setSliceList(..))
//	 * @return
//	 * @throws MTBOperatorException
//	 */
//	public Vector<MTBPoint3D> getSliceList() throws MTBOperatorException {
//		return (Vector<MTBPoint3D>)this.getParameter("sliceList");
//	}

	@Override
	public void validateCustom() throws MTBOperatorException {
		if (!(new File(this.getFileName())).exists()) {
			throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED, "MTBImageReader.validateCustom():" +
					" Cannot read file '"+this.getFileName()+"'. File does not exists.");
		}
	}
	
	@Override
	protected void operate() throws MTBOperatorException,
			MTBProcessingDAGException {
		
		try {
			this.setResultImage(this.readMTBImage(this.getFileName()));
		} catch (IOException e) {
			e.printStackTrace();
			throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageReader.operate(): " +
					"IOException thrown in readMTBImage method: " + e.getMessage());		
		}
	}
	
	public MTBImage readMTBImage(String filename) throws MTBOperatorException, IOException {

		int sizeX = 0, sizeY = 0, sizeZ = 0, sizeT = 0, sizeC = 0;
		double ssX = 1, ssY = 1, ssZ = 1, ssT = 1;
		String uX = "", uY = "", uZ = "", uT = "";
		
		try {
			ImgPropsReader ipr = new ImgPropsReader(filename);
			
			sizeX = ipr.sizeX;
			sizeY = ipr.sizeY;
			sizeZ = ipr.sizeZ;
			sizeT = ipr.sizeT;
			sizeC = ipr.sizeC;
			ssX = ipr.stepsizeX;
			ssY = ipr.stepsizeY;
			ssZ = ipr.stepsizeZ;
			ssT = ipr.stepsizeT;
			uX = ipr.unitX;
			uY = ipr.unitY;
			uZ = ipr.unitZ;
			uT = ipr.unitT;
			
		} catch (IllegalArgumentException e) {
			
		}

		
        // Get a TIFF reader
        ImageReader reader = new TIFFImageReader(null);

        // Create the read parameter object
        TIFFImageReadParam readParam = new TIFFImageReadParam();
        
        ImageInputStream input = ImageIO.createImageInputStream(new File(filename));
        reader.setInput(input);
        
        
        // get the number of images in the input stream of the reader
        int nImages = -1;
        try {
        	nImages = reader.getNumImages(true);
        }
        catch (IllegalStateException e) {
        	nImages = reader.getNumImages(false);
        }
        
		DataBuffer dbuffer = null;
		RenderedImage rimg = null;
		MTBImage outimg = null;
        MTBImageType outtype = null;
        int rsizeX, rsizeY;
        
        if (nImages >= 1) {
        	if (this.getVerbose()) {
	        	Iterator<ImageTypeSpecifier> titer = reader.getImageTypes(0);
	        	ImageTypeSpecifier spec;
	        	while (titer.hasNext()) {
	        		spec = titer.next();
	        		String s = "TIFF Image properties: ";
	        		s += "Components: " + spec.getNumComponents() + "\n";
	        		s += "Bands: " + spec.getNumBands() + "\n";
	        		
	        		for (int b = 0; b < spec.getNumBands(); b++)
	        			s += "Bits per Band " + b + ": "+ spec.getBitsPerBand(b) + "\n";
	        		
	        		
	        		System.out.println(s);
	        	}
        	}
        	
        	
        	rimg = reader.readAsRenderedImage(0, readParam);
        	outtype = this.mtbtype(rimg);
        	
        	if (outtype == null)
        		throw new IOException("MTBImageReader.readMTBImage(.): Unknown image data types are present in the TIFF file. Cannot create an MTBImage.");
        	
    		rsizeX = rimg.getWidth();
    		rsizeY = rimg.getHeight();

			if (sizeX != 0) {
    			// if image dimension properties were read from history file, but do not match
    			// the image dimension properties of the present TIFF file, throw an exception
    			
    			if (rsizeX != sizeX || rsizeY != sizeY || sizeZ*sizeT*sizeC != nImages)
    				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImagerReader.readMTBImage(.): " +
    						" Image dimension properties from .mph file do not match present image dimensions.");
    			
    		}
    		else {
    			// if image dimension properties could not be read from history file
    			// create a z-stack with 'nImages' slices
    			
    			sizeX = rsizeX;
    			sizeY = rsizeY;
    			sizeZ = nImages;
    			sizeT = 1;
    			sizeC = 1;
    		}
			
			outimg = MTBImage.createMTBImage(sizeX, sizeY, sizeZ, sizeT, sizeC, outtype);
        	
        	
        	// set image properties
        	outimg.setStepsizeX(ssX); outimg.setStepsizeY(ssY); outimg.setStepsizeZ(ssZ); outimg.setStepsizeT(ssT); 
        	outimg.setUnitX(uX); outimg.setUnitY(uY); outimg.setUnitZ(uZ); outimg.setUnitT(uT);
        	
        	// read image history
        	outimg.readHistory(filename);
        	
        	// read image data
    		int i = 0;
    		
    		do {
    			outimg.setActualSliceIndex(i);
    			dbuffer = rimg.getData().getDataBuffer();
    			if (outtype != this.mtbtype(rimg))
    				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, 
    						"MTBImageReader.readMTBImage(.): Different image types are present in the TIFF file. Cannot create an MTBImage.");
    			
    			if (rimg.getWidth() != sizeX || rimg.getHeight() != sizeY) {
    				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, 
    						"MTBImageReader.readMTBImage(.): Images of different size are present in the TIFF file. Cannot create an MTBImage.");
    			}
    			
    			
    			if (outtype == MTBImageType.MTB_RGB) {
    				// red channel
	    			for (int y = 0; y < sizeY; y++)
	    				for (int x = 0; x < sizeX; x++)
	    					((MTBImageRGB)outimg).putValueR(x, y, rimg.getData().getSample(x, y, 0));//dbuffer.getElem(0, y*sizeX + x));
    				// green channel
	    			for (int y = 0; y < sizeY; y++)
	    				for (int x = 0; x < sizeX; x++)
	    					((MTBImageRGB)outimg).putValueG(x, y, rimg.getData().getSample(x, y, 1));//dbuffer.getElem(0, y*sizeX + x));
    				// blue channel
	    			for (int y = 0; y < sizeY; y++)
	    				for (int x = 0; x < sizeX; x++)
	    					((MTBImageRGB)outimg).putValueB(x, y, rimg.getData().getSample(x, y, 2));//dbuffer.getElem(0, y*sizeX + x));			
    			}
    			else if (outtype.ordinal() <= 2) {
    				// integer types
	    			for (int y = 0; y < sizeY; y++)
	    				for (int x = 0; x < sizeX; x++)
	    					outimg.putValueInt(x, y, dbuffer.getElem(y*sizeX + x));
    			}
    			else {
    				// floating point types
	    			for (int y = 0; y < sizeY; y++)
	    				for (int x = 0; x < sizeX; x++)
	    					outimg.putValueDouble(x, y, dbuffer.getElemDouble(y*sizeX + x));
    			}
	
    			i++;
    			try {
    				rimg = reader.readAsRenderedImage(i, readParam);
    			} catch (IndexOutOfBoundsException e) {
    				i = nImages;
    			}
    			
    		} while (i < nImages);
    		
    		return outimg;
        	
        }
        else {
        	throw new IOException("MTBImageReader.readMTBImage(.): Cannot retrieve the number of images in the TIFF file. No MTBImage created.");
        }

	}
	
	

	protected MTBImageType mtbtype(RenderedImage rimg) {
		
		int type = rimg.getSampleModel().getDataType();
		int bands = rimg.getSampleModel().getNumBands();
		
		if (type == DataBuffer.TYPE_BYTE && bands == 1)
			return MTBImageType.MTB_BYTE;
		else if (type == DataBuffer.TYPE_BYTE && bands == 3) 
			return MTBImageType.MTB_RGB;
		else if (type == DataBuffer.TYPE_USHORT && bands == 1)
			return MTBImageType.MTB_SHORT;
		else if (type == DataBuffer.TYPE_INT && bands == 1)
			return MTBImageType.MTB_INT;
		else if (type == DataBuffer.TYPE_FLOAT && bands == 1)
			return MTBImageType.MTB_FLOAT;
		else if (type == DataBuffer.TYPE_DOUBLE && bands == 1)
			return MTBImageType.MTB_DOUBLE;
		else
			return null;
	}
	
	private class ImgPropsReader {
		
		int sizeX = -1, sizeY = -1, sizeZ = -1, sizeT = -1, sizeC = -1;
		double stepsizeX = 1, stepsizeY = 1, stepsizeZ = 1, stepsizeT = 1;
		String unitX = "", unitY = "", unitZ = "", unitT = "";
		
		ImgPropsReader(String filename) throws IllegalArgumentException {
			// dummy image to read history graph
			MTBImage img = MTBImage.createMTBImage(1, 1, 1, 1, 1, MTBImageType.MTB_BYTE);

			img.readHistory(filename);

			
			// the history graph also contains image properties
			GraphmlType gml = img.getGraphmlHistory();
			
			// read image properties
			if (gml != null) {
				
				this.fetchImgProps(gml);
			}
			else {
				throw new IllegalArgumentException("ReaderMTBImage.ImgPropsReader.ImgPropsReader(.): No Graphml history available, cannot fetch image properties.");
			}
		}
		
		private void fetchImgProps(GraphmlType gml) {
			GraphType g = gml.getGraphArray()[0];
			NodeType n = this.getImgPropsNode(g);
			if (n == null)
				throw new IllegalArgumentException("ReaderMTBImage.ImgPropsReader.fetchImgProps(.): Can't find image node, cannot fetch image properties.");
			
			if (this.isValidImgPropsNode(n)) {
			
				PropertyType[] parray = this.getPropsArray(n);
						
				for (int i = 0; i < parray.length; i++) {
					if (parray[i].getKey().compareTo("SizeX") == 0)
						this.sizeX = Integer.parseInt(parray[i].getValue());
					else if (parray[i].getKey().compareTo("SizeY") == 0)
						this.sizeY = Integer.parseInt(parray[i].getValue());
					else if (parray[i].getKey().compareTo("SizeZ") == 0)
						this.sizeZ = Integer.parseInt(parray[i].getValue());
					else if (parray[i].getKey().compareTo("SizeT") == 0)
						this.sizeT = Integer.parseInt(parray[i].getValue());
					else if (parray[i].getKey().compareTo("SizeC") == 0)
						this.sizeC = Integer.parseInt(parray[i].getValue());
					else if (parray[i].getKey().compareTo("StepsizeX") == 0)
						this.stepsizeX = Double.parseDouble(parray[i].getValue());
					else if (parray[i].getKey().compareTo("StepsizeY") == 0)
						this.stepsizeY = Double.parseDouble(parray[i].getValue());
					else if (parray[i].getKey().compareTo("StepsizeZ") == 0)
						this.stepsizeZ = Double.parseDouble(parray[i].getValue());
					else if (parray[i].getKey().compareTo("StepsizeT") == 0)
						this.stepsizeT = Double.parseDouble(parray[i].getValue());
					else if (parray[i].getKey().compareTo("UnitX") == 0)
						this.unitX = parray[i].getValue();
					else if (parray[i].getKey().compareTo("UnitY") == 0)
						this.unitY = parray[i].getValue();
					else if (parray[i].getKey().compareTo("UnitZ") == 0)
						this.unitZ = parray[i].getValue();
					else if (parray[i].getKey().compareTo("UnitT") == 0)
						this.unitT = parray[i].getValue();
				}
			}
		
		}
		
		
		private NodeType getImgPropsNode(GraphType graph) {
			NodeType[] nodes = graph.getNodeArray();
			
			if (nodes != null) {
				if (nodes[1].getId().matches("MTBDAG[0-9]+:DataPort[0-9]+")) {
					return nodes[1];
				}
				else if (nodes[1].getId().matches("MTBDAG[0-9]+:OutputPort[0-9]+.*")) {
					return nodes[1];
				}
				else {
					return this.getImgPropsNode(nodes[1].getGraph());
				}
				
			}
			else {
				return null;
			}
			
		}
			
		private boolean isValidImgPropsNode(NodeType node) {
			DataType[] data = node.getDataArray();
			PropertyType[] props = null;
			
			for (int i = 0; i < data.length; i++) {
				if (data[i].getKey().compareTo("mtbProps") == 0) {
					props = data[i].getProperties().getPropertyArray();
					break;
				}
			}
			if (props != null) {
				for (int i = 0; i < props.length; i++) {
					if (props[i].getKey().compareTo("SizeT") == 0) {
						return true;
					}
				}
			}
			
			return false;
		}
		
		private PropertyType[] getPropsArray(NodeType node) {
			DataType[] data = node.getDataArray();
			
			for (int i = 0; i < data.length; i++) {
				if (data[i].getKey().compareTo("mtbProps") == 0) {
					return data[i].getProperties().getPropertyArray();
				}
			}
			
			return null;
		}
		
	}

}
