/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.io.files;

import it.geosolutions.imageio.plugins.tiff.TIFFImageWriteParam;
import it.geosolutions.imageioimpl.plugins.tiff.TIFFImageWriter;

import java.awt.Point;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.color.ICC_ColorSpace;
import java.awt.color.ICC_Profile;
import java.awt.image.BandedSampleModel;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriter;
import javax.imageio.stream.ImageOutputStream;

import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageRGB;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * Operator for writing MTBImages to disk. 
 * The image is written in TIFF format.
 * The history graph is stored automatically with the same file basename and .mph extension
 * @author Oliver Gress
 *
 */
public class WriterMTBImage extends MTBOperator {

	/**
	 * Constructor.
	 * @param img image to store
	 * @param filename image filename (in TIFF format)
	 * @throws MTBOperatorException
	 */
	public WriterMTBImage(MTBImage img, String filename) throws MTBOperatorException {
		this.setInputImage(img);
		this.setFileName(filename);
		
		this.collectArguments();
	}
	
	
	
	@Override
	protected void collectArguments() throws MTBOperatorException {
		super.collectArguments();


		MTBOpArgumentDescriptor[] inputDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
		    "inputImage", MTBImage.class, "Image to save", true, null) };

		
		MTBOpArgumentDescriptor[] outputDescriptors = null;

		
		MTBOpArgumentDescriptor[] parameterDescriptors = new MTBOpArgumentDescriptor[] {
		    new MTBOpArgumentDescriptor("fileName", java.lang.String.class,
		        "Image file name", true, null),
		    new MTBOpArgumentDescriptor("overwrite", java.lang.Boolean.class,
		        "Overwrite permission flag", true, false) };
		
		
		MTBOpArgumentDescriptor[] supplementalDescriptor = null;
		//new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
	//	    "verbose", java.lang.Boolean.class,
	//	    "Verbose mode", true, true) };

		
		this.addDescriptors(parameterDescriptors, inputDescriptors, outputDescriptors,
		    supplementalDescriptor);
	}
	
	/**
	 * Set image to store to disk
	 * @param img
	 * @throws MTBOperatorException
	 */
	public void setInputImage(MTBImage img) throws MTBOperatorException {
		this.setInput("inputImage", img);
	}
	
	/**
	 * Get image that has to be stored to disk
	 * @return
	 * @throws MTBOperatorException
	 */
	public MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)this.getInput("inputImage");
	}
	
	/**
	 * Get image filename
	 * @param filename
	 * @throws MTBOperatorException
	 */
	public void setFileName(String filename) throws MTBOperatorException {
		this.setParameter("fileName", filename);
	}
	
	/**
	 * Set image filename
	 * @return
	 * @throws MTBOperatorException
	 */
	public String getFileName() throws MTBOperatorException {
		return (String)this.getParameter("fileName");
	}
	
	/**
	 * Set the overwrite permission flag
	 * @param overwrite
	 * @throws MTBOperatorException
	 */
	public void setOverwrite(boolean overwrite) throws MTBOperatorException {
		this.setParameter("overwrite", overwrite);
	}
	
	/**
	 * Get the overwrite permission flag
	 * @return
	 * @throws MTBOperatorException
	 */
	public boolean getOverwrite() throws MTBOperatorException {
		return (Boolean)this.getParameter("overwrite");
	}

	@Override
	public void validateCustom() throws MTBOperatorException {
		if (this.getOverwrite() == false && (new File(this.getFileName())).exists()) {
			throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED, "MTBImageWriter.validateCustom():" +
					" Cannot write to file '"+this.getFileName()+"'. File exists and overwrite is forbidden.");
		}
	}
	
	@Override
	protected void operate() throws MTBOperatorException,
			MTBProcessingDAGException {
		
		try {
			this.writeMTBImage(this.getFileName(), this.getInputImage());
		} catch (IOException e) {
			e.printStackTrace();
			throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageWriter.operate(): " +
					"IOException thrown in writeMTBImage method: " + e.getMessage());		
		}
	}
	
	protected void writeMTBImage(String filename, MTBImage img) 
					throws MTBOperatorException, IOException, MTBProcessingDAGException {
		
		int sizeX = img.getSizeX();
		int sizeY = img.getSizeY();
		int sizeStack = img.getSizeStack();
		
		
        // Get a TIFF writer and set its output.
//       Iterator writers = ImageIO.getImageWritersByFormatName("tiff");
//       if(writers == null || !writers.hasNext()) {
//            throw new RuntimeException("MTBImageWriter.writeMTBImage(.): No TIFF writers available. Make sure to link the JAI ImageIO library.");
//        }
//      while (writers.hasNext())
//     	System.out.println(writers.next().toString());
       
		// Create new TIFF image writer
        ImageWriter writer = new TIFFImageWriter(null); //(ImageWriter)writers.next();
       // System.out.println(writer.toString());

        // Create the write parameter object
        TIFFImageWriteParam writeParam = new TIFFImageWriteParam(null);
        
        ImageOutputStream output = ImageIO.createImageOutputStream(new File(filename));
        writer.setOutput(output);
        
        // required objects for RenderedImage construction for input to the writer
		WritableRaster raster = null;
		SampleModel sm = null;
		ColorModel cm = null;
		RenderedImage rimg = null;
		
		
		if (img.getType() != MTBImageType.MTB_RGB) {
		// write gray images
			
			int dbuffertype = -1;
			
			// data buffer for file output
			DataBuffer d = null;
			
			// select DataBuffer type and create corresponding DataBuffer
			if (img.getType() == MTBImageType.MTB_BYTE) {
				dbuffertype = DataBuffer.TYPE_BYTE;
				d = new DataBufferByte(sizeY*sizeX, 1);	
			}
			else if (img.getType() == MTBImageType.MTB_SHORT) {
				dbuffertype = DataBuffer.TYPE_USHORT;
				d = new DataBufferUShort(sizeY*sizeX, 1);	
			}
			else if (img.getType() == MTBImageType.MTB_INT) {
				dbuffertype = DataBuffer.TYPE_INT;
				d = new DataBufferInt(sizeY*sizeX, 1);
			}
			else if (img.getType() == MTBImageType.MTB_FLOAT) {
				dbuffertype = DataBuffer.TYPE_FLOAT;
				d = new DataBufferFloat(sizeY*sizeX, 1);
			}
			else if (img.getType() == MTBImageType.MTB_DOUBLE) {
				dbuffertype = DataBuffer.TYPE_DOUBLE;
				d = new DataBufferDouble(sizeY*sizeX, 1);
			}
			else {
				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED,
								"MTBImageWriter.writeMTBImage(.): Unknown MTBImageType: " + img.getType());
			}
			
			// simple sample model with one 'double' value per pixel per slice
			sm = new BandedSampleModel(dbuffertype, sizeX, sizeY, 1);

			// gray color model
			cm = new ComponentColorModel(new ICC_ColorSpace(ICC_Profile.getInstance(ColorSpace.CS_GRAY)),
					false, false, Transparency.OPAQUE, dbuffertype);

			
			// write every slice to the image file
			int idx = img.getActualSliceIndex();
			for (int i = 0; i < sizeStack; i++) {
				img.setActualSliceIndex(i);
			
				// test if further images can be inserted to the image file
		        if (i > 0 && !writer.canInsertImage(-1)) {
		        	throw new IOException("MTBImageWriter.writeMTBImage(.): Available image writer cannot handle multiple images in one file.");
		        }
		        
				// write data to data buffer
				for (int y = 0; y < sizeY; y++)
					for (int x = 0; x < sizeX; x++)
						d.setElemDouble(y*sizeX + x, img.getValueDouble(x, y));
				
				// create java image for ImageIO writing
				raster = Raster.createWritableRaster(sm, d, new Point(0,0));
				rimg = new BufferedImage(cm, raster, false, null);
				
				// write
				if (i == 0)
					writer.write(null, new IIOImage(rimg, null, null), writeParam);
				else
					writer.writeInsert(-1, new IIOImage(rimg, null, null), writeParam);
			}
			// restore actual slice index
			img.setActualSliceIndex(idx);	

		} 
		else {
		// write RGB color image (MTBImageType.MTB_RGB
			
			DataBuffer d = new DataBufferDouble(sizeY*sizeX, 3);
			
			sm = new BandedSampleModel(DataBuffer.TYPE_BYTE, sizeX, sizeY, 3);
			cm = new ComponentColorModel(new ICC_ColorSpace(ICC_Profile.getInstance(ColorSpace.CS_LINEAR_RGB)),
					false, false, Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
			
			MTBImageRGB rgbimg = (MTBImageRGB)img;
			
			int idx = rgbimg.getActualSliceIndex();
			
			for (int i = 0; i < sizeStack; i++) {
				rgbimg.setActualSliceIndex(i);
			
		        if (i > 0 && !writer.canInsertImage(-1)) {
		        	throw new IOException("MTBImageWriter.writeMTBImage(.): Available image writer cannot handle multiple images in one file.");
		        }
				
				
				// red channel
				for (int y = 0; y < sizeY; y++)
					for (int x = 0; x < sizeX; x++)
						d.setElemDouble(0, y*sizeX + x, rgbimg.getValueR(x, y));
				
				// green channel
				for (int y = 0; y < sizeY; y++)
					for (int x = 0; x < sizeX; x++)
						d.setElemDouble(1, y*sizeX + x, rgbimg.getValueG(x, y));
				
				// blue channel
				for (int y = 0; y < sizeY; y++)
					for (int x = 0; x < sizeX; x++)
						d.setElemDouble(2, y*sizeX + x, rgbimg.getValueB(x, y));
				
				raster = Raster.createWritableRaster(sm, d, new Point(0,0));
				rimg = new BufferedImage(cm, raster, false, null);
				
				if (i == 0)
					writer.write(null, new IIOImage(rimg, null, null), writeParam);
				else
					writer.writeInsert(-1, new IIOImage(rimg, null, null), writeParam);
			}
			rgbimg.setActualSliceIndex(idx);	
		}
		
		
		img.writeHistory(filename);
	}

}
