/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.morphology;

import ij.process.*;
import java.awt.geom.*;
import java.util.*;

import de.unihalle.informatik.MiToBo.datatypes.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.*;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents.ContourType;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.*;

/**
 * This class provides methods for post-processing components in binary images
 * with regard to their morphology.
 * <p>
 * Thresholding algorithms yield binary images with distinct connected
 * components of foreground pixels, i.e. regions. The routines of this class
 * provide functionality for post-processing such binary (unlabeled) component
 * images. Examplary post-processing steps might include linking/merging
 * adjacent components or changing a component's morphological shape.
 * 
 * @author moeller
 */
public class ComponentPostprocess extends MTBOperator {

		/**
		 * Processing mode identifiers.
		 * 
		 * @author moeller
		 */
		public static enum ProcessMode {
				/**
				 * Delete components below a size threshold from the image.
				 */
				ERASE_SMALL_COMPS,
				/**
				 * Link adjacent components with a line if distance below threshold.
				 */
				LINK_ADJ_COMPS,
				/**
				 * Perform Voronoi expansion for pixels close to components.
				 */
				VORONOI_EXPAND,
				/**
				 * Delete components with a rounded shape instead of an sustained shape.
				 */
				 // TODO by Danny
				ERASE_ROUND_COMPS
		}

		protected void collectArguments()  throws MTBOperatorException {
				super.collectArguments();

				MTBOpArgumentDescriptor[] inputDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
				    "inImg", MTBImageByte.class, "Input image", true, null) };

				MTBOpArgumentDescriptor[] outputDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
				    "resultImg", MTBImageByte.class, "Result image", true, null) };

				MTBOpArgumentDescriptor[] parameterDescriptors = new MTBOpArgumentDescriptor[] {
				    new MTBOpArgumentDescriptor("processMode", ProcessMode.class,
				        "Process Modus", true, null),
				    new MTBOpArgumentDescriptor("minCompSize", java.lang.Integer.class,
				        "Minimal Component Size", true, null),
				    new MTBOpArgumentDescriptor("maxCompDist", java.lang.Integer.class,
				        "Maximal Component Distance", true, null),
				    new MTBOpArgumentDescriptor("maxVoronoiExpDist",
				        java.lang.Integer.class, "Maximal Voronoi Expand Distance", true,
				        null),
				    new MTBOpArgumentDescriptor("DiagonalNeighbors",
				        java.lang.Boolean.class,
				        "true for 8-neighborhood, false for 4-neighborhood", true, true) };

				MTBOpArgumentDescriptor[] supplementalDescriptors = null;

				addDescriptors(parameterDescriptors, inputDescriptors, outputDescriptors,
				    supplementalDescriptors);
		}

		public ComponentPostprocess() throws MTBOperatorException {
				// nothing to do here
		}

		/**
		 * Default constructor. A new empty meta data object is initialized.
		 * 
		 * @param inImg
		 *          Image to work on.
		 * @param pm
		 *          Process mode of operator.
		 * @param minCompSize
		 *          Minimal region size for eliminating small comps.
		 * @param maxCompDist
		 *          Maximal region distance in linking.
		 * @param maxVoronoiExpandDist
		 *          Max. distance of pixel considered in Voronoi expansion.
		 * 
		 *          Note: Depending on the process mode each time only one of the
		 *          parameters is actually in use.
		 * @throws MTBProcessingDAGException
		 */
		public ComponentPostprocess(MTBImageByte inImg, ProcessMode pm,
		    int minCompSize, int maxCompDist, int maxVoronoiExpandDist)
		    throws MTBOperatorException {
				this.setInput("inImg", inImg);
				this.setParameter("processMode", pm);
				this.setParameter("minCompSize", new Integer(minCompSize));
				this.setParameter("maxCompDist", new Integer(maxCompDist));
				this.setParameter("maxVoronoiExpDist", new Integer(maxVoronoiExpandDist));
		}

		/**
		 * Get reference to the current input image.
		 * 
		 * @return Input image to work on.
		 * @throws MTBProcessingDAGException
		 */
		public MTBImageByte getInputImage() throws MTBOperatorException {
				return (MTBImageByte) (this.getInput("inImg"));
		}

		/**
		 * Get current masksize.
		 * 
		 * @return Masksize of dilation mask.
		 * @throws MTBProcessingDAGException
		 */
		public ProcessMode getProcessMode() throws MTBOperatorException {
				return (ProcessMode) (this.getParameter("processMode"));
		}

		/**
		 * Returns the minimal valid component size for component removal.
		 * 
		 * @throws MTBProcessingDAGException
		 */
		public int getMinimalComponentSize() throws MTBOperatorException {
				return ((Integer) (this.getParameter("minCompSize"))).intValue();
		}

		/**
		 * Returns the maximal component distance for linking.
		 * 
		 * @throws MTBProcessingDAGException
		 */
		public int getMaximalComponentDist() throws MTBOperatorException {
				return ((Integer) (this.getParameter("maxCompDist"))).intValue();
		}

		/**
		 * Returns the maximal distance of pixels considered in Voronoi expansion.
		 * 
		 * @throws MTBProcessingDAGException
		 */
		public int getMaxVoronoiExpandDist() throws MTBOperatorException {
				return ((Integer) (this.getParameter("maxVoronoiExpDist"))).intValue();
		}

		/**
		 * Get value of Parameter argument DiagonalNeighbors.
		 * 
		 * @return value of DiagonalNeighbors
		 */
		public java.lang.Boolean getDiagonalNeighbors() throws MTBOperatorException {
				return (java.lang.Boolean) (this.getParameter("DiagonalNeighbors"));
		}

		/**
		 * Set value of Parameter argument DiagonalNeighbors.
		 * 
		 * @param value
		 *          New value for DiagonalNeighbors
		 */
		public void setDiagonalNeighbors(java.lang.Boolean value)
		    throws MTBOperatorException {
				this.setParameter("DiagonalNeighbors", value);
		}

		/**
		 * Get the result image after applying operator. Attention, reference might be
		 * null.
		 * 
		 * @return Reference to result image.
		 * @throws MTBProcessingDAGException
		 */
		public MTBImageByte getResultImage() throws MTBOperatorException {
				return (MTBImageByte) (this.getOutput("resultImg"));
		}

		/**
		 * Set the result image.
		 * 
		 * @param rimage
		 *          Result image.
		 * @throws MTBProcessingDAGException
		 */
		protected void setResultImage(MTBImageByte rimage)
		    throws MTBOperatorException {
				this.setOutput("resultImg", rimage);
		}

		// input image width and height
		private int width;
		private int height;

		// temporary variables needed during internal calculations
		private Vector<Vector<Point2D.Double>> endPointHash;

		@Override
		protected void operate() throws MTBOperatorException,
		    MTBProcessingDAGException {

				MTBImageByte result = null;
				switch (this.getProcessMode()) {
				case ERASE_SMALL_COMPS:
						result = this.EraseSmallComponents(this.getInputImage(), this
						    .getMinimalComponentSize());
						break;
				case LINK_ADJ_COMPS:
						result = this.LinkAdjacentComponents(this.getInputImage(), this
						    .getMaximalComponentDist());
						break;
				case VORONOI_EXPAND:
						result = this.VoronoiExpandComponents(this.getInputImage(), this
						    .getMaxVoronoiExpandDist());
						break;
				}
				this.setResultImage(result);

				if (result == null)
						throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED,
						    "ComponentPostprocess.operate() failed: " + "Unknown process mode "
						        + this.getProcessMode());
		}

		/**
		 * This function removes foreground components from a binary image that have a
		 * size below the given threshold. Suitable for noise reduction in binary
		 * images.
		 * 
		 * @param binIP
		 *          binary input image
		 * @param minComponentSize
		 *          minimum size of valid components
		 * @return image with too small components erased
		 * @throws MTBOperatorException
		 */
		protected MTBImageByte EraseSmallComponents(MTBImageByte binIP,
		    int minComponentSize) throws MTBOperatorException,
		    MTBProcessingDAGException {

				this.width = binIP.getSizeX();
				this.height = binIP.getSizeY();

				// initialize result image
				MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(
				    this.width, this.height, 1, 1, 1, MTBImageType.MTB_BYTE);
				resultImage.fillWhite();

				// copy input image data
				resultImage.setImagePart(binIP, 0, 0, 0, 0, 0);

				// label components in binary image
				LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				    binIP, getDiagonalNeighbors());
				regionLabler.runOp(this);
				MTBRegion2DSet binRegions = regionLabler.getResultingRegions();

				for (int i = 0; i < binRegions.size(); ++i) {
						MTBRegion2D r = binRegions.elementAt(i);
						if (r.getArea() < minComponentSize) {
								// iterate over all region points...
								Vector<Point2D.Double> points = r.getPoints();
								for (int j = 0; j < points.size(); ++j) {

										int x = (int) points.elementAt(j).x;
										int y = (int) points.elementAt(j).y;

										resultImage.putValueInt(x, y, 0);
								}
						}
				}
				return resultImage;
		}

		/**
		 * Function for linking adjacent components/regions in a binary image.
		 * <p>
		 * This function processes a binary image with foreground objects being white.
		 * Based on an initial component labeling step and subsequent skeletonization
		 * of resulting components, adjacent components are linked by a line of width
		 * one pixel in the result image.
		 * <p>
		 * The criterion for two components to be adjacent is defined based on the
		 * distance of pairs of skeleton endpoints of both components. If there exists
		 * at least one pair of endpoints between two components with a distance below
		 * the given threshold, the components are linked. If there is more than one
		 * pair of such points, the pair with the smallest distance is selected to be
		 * linked to each other.
		 * 
		 * @param binImage
		 *          input binary image with regions
		 * @return binary image where adjacent components are linked by a line of
		 *         width 1 pixel
		 * @throws MTBOperatorException
		 */
		protected MTBImageByte LinkAdjacentComponents(MTBImageByte binIP,
		    int maxCompDist) throws MTBOperatorException, MTBProcessingDAGException {

				this.width = binIP.getSizeX();
				this.height = binIP.getSizeY();

				// initialize result image
				MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(
				    this.width, this.height, 1, 1, 1, MTBImageType.MTB_BYTE);
				resultImage.fillWhite();

				// copy input image data
				resultImage.setImagePart(binIP, 0, 0, 0, 0, 0);

				// label components in binary image
				LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				    binIP, getDiagonalNeighbors());
				regionLabler.runOp(this);
				MTBRegion2DSet binRegions = regionLabler.getResultingRegions();

				// calculate skeletons of initial binary regions,
				// skeletonize function requires inverse input image (object = black)
				ByteProcessor bP = new ByteProcessor(this.width, this.height);
				for (int y = 0; y < this.height; ++y) {
						for (int x = 0; x < this.width; ++x) {
								if (binIP.getValueInt(x, y) > 0)
										bP.putPixel(x, y, 0);
								else
										bP.putPixel(x, y, 255);
						}
				}
				BinaryProcessor bbP = new BinaryProcessor(bP);
				bbP.skeletonize();

				// array where endpoints are marked
				int endPointArray[][] = new int[this.width][this.height];

				// pseudo-hashtable with all endpoints of binary regions:
				// remember the endpoints of all regions, indexed by region index
				this.endPointHash = new Vector<Vector<Point2D.Double>>();
				for (int i = 0; i < binRegions.size(); ++i) {
						this.endPointHash.add(new Vector<Point2D.Double>());
				}

				// find all endpoints of all binary regions
				for (int i = 0; i < binRegions.size(); ++i) {
						MTBRegion2D r = binRegions.elementAt(i);
						Vector<Point2D.Double> points = r.getPoints();

						// iterate over all region points...
						for (int j = 0; j < points.size(); ++j) {

								int x = (int) points.elementAt(j).x;
								int y = (int) points.elementAt(j).y;

								// found a point of the skeleton?
								if (bbP.getPixel(x, y) == 0) {

										// if so, count neighboring skeleton points
										int nCount = 0;
										for (int dx = -1; dx <= 1; dx++) {
												for (int dy = -1; dy <= 1; ++dy) {
														if (dx == 0 && dy == 0)
																continue;
														if (bbP.getPixel(x + dx, y + dy) == 0)
																nCount++;
												}
										}
										// if there is only one neighbor in 8-NB, (x,y) is endpoint
										if (nCount == 1) {
												endPointArray[y][x] = 1;
												// binEndPoints.add(new Point2D.Double(x,y));

												// put endpoint in hash table
												this.endPointHash.elementAt(i).add(new Point2D.Double(x, y));
										}
								}
						}
				} // end of for-loop: all binary regions

				// iterate over all regions and link them to neighbor regions
				for (int i = 0; i < binRegions.size() - 1; ++i) {

						// get endpoints of this region
						Vector<Point2D.Double> regionEndPoints = this.endPointHash.elementAt(i);

						// find close endpoints in neighboring regions
						for (int j = i + 1; j < binRegions.size(); ++j) {

								// get endpoints of neighbor region
								Vector<Point2D.Double> neighborEndPoints = this.endPointHash
								    .elementAt(j);

								double regionDist = maxCompDist;
								int closestPointID_region = -1;
								int closestPointID_neighbor = -1;

								for (int m = 0; m < regionEndPoints.size(); ++m) {
										double rx = regionEndPoints.elementAt(m).x;
										double ry = regionEndPoints.elementAt(m).y;

										for (int n = 0; n < neighborEndPoints.size(); ++n) {
												double nx = neighborEndPoints.elementAt(n).x;
												double ny = neighborEndPoints.elementAt(n).y;

												double dist = (rx - nx) * (rx - nx) + (ry - ny) * (ry - ny);

												// accept endpoints with distance below threshold
												if (dist < regionDist) {
														closestPointID_region = m;
														closestPointID_neighbor = n;
														regionDist = dist;
												}
										}
								}
								// check if regions are close to each other
								if (regionDist < maxCompDist) {
										// if so, draw line between the two endpoints
										Point2D.Double regPoint = regionEndPoints
										    .elementAt(closestPointID_region);
										Point2D.Double neighbPoint = neighborEndPoints
										    .elementAt(closestPointID_neighbor);
										resultImage.drawLine2D((int) regPoint.x, (int) regPoint.y,
										    (int) neighbPoint.x, (int) neighbPoint.y, 255);
								}
						} // end of for-loop: iterate over all pairs of region endpoints
				} // end of for-loop: iterate over all regions

				return resultImage;

		} // end of function: LinkRegions(...)

		protected MTBImageByte VoronoiExpandComponents(MTBImageByte binImg,
		    int maxDist) throws MTBOperatorException, MTBProcessingDAGException {

				// ImageProcessor binIP = binImg.getImagePlus().getProcessor();

				ImgDilate dilOp = new ImgDilate(binImg, maxDist * 2);
				dilOp.runOp(this);
				MTBImage dilateBinImg = dilOp.getResultImage();

				// pixels in dilated area
				MTBImage relPixels = MTBImage.createMTBImage(binImg.getSizeX(), binImg
				    .getSizeY(), 1, 1, 1, MTBImageType.MTB_BYTE);
				for (int y = 0; y < binImg.getSizeY(); ++y) {
						for (int x = 0; x < binImg.getSizeX(); ++x) {
								if (dilateBinImg.getValueInt(x, y) > 0 && binImg.getValueInt(x, y) == 0)
										relPixels.putValueInt(x, y, 255);
						}
				}

				// component labeling
				LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				    binImg, getDiagonalNeighbors());
				regionLabler.runOp(this);
				MTBRegion2DSet comps = regionLabler.getResultingRegions();

				/**
				 * to do: Change ImagePlus and ImageProcessors to MTBImage objects. new
				 * version: changed by danny, using new ContourOnLabeledComponents class
				 */
				MTBImageByte binImgByte = (MTBImageByte) binImg.convertType(
				    MTBImageType.MTB_BYTE, false);
				ContourOnLabeledComponents clc = new ContourOnLabeledComponents(binImgByte,
				    comps, ContourType.OUTER_CONTOUR, 1);
				clc.runOp(this);
				MTBContour2DSet conts = clc.getContours();

				/**
				 * old version: changed by danny
				 */
				// Vector<Contour2D> conts=
				// Contour_on_Labeled_Components.get_ContoursOnRegions(binIP,comps);

				// check each pixel in the image
				for (int y = 0; y < binImg.getSizeY(); ++y) {
						for (int x = 0; x < binImg.getSizeX(); ++x) {

								// pixel is not new, i.e. was there before -> ignore it
								if (relPixels.getValueInt(x, y) == 0)
										continue;

								double minDist_l = 100000000;
								int minID = -1;

								// alle Konturen
								minDist_l = 100000000;
								for (int i = 0; i < conts.size(); ++i) {
										MTBContour2D c = conts.elementAt(i);
										Vector<Point2D.Double> pc = c.getPoints();

										// alle Punkte der Kontur
										double c_dist = 100000000;
										for (int k = 0; k < pc.size(); ++k) {
												Point2D.Double pk = pc.elementAt(k);
												double d = Math.sqrt((x - pk.getX()) * (x - pk.getX())
												    + (y - pk.getY()) * (y - pk.getY()));
												if (d < c_dist) {
														c_dist = d;
												}
										}
										if (c_dist < minDist_l) {
												minDist_l = c_dist;
												minID = i;
										}
								}
								if (minDist_l < maxDist && minID != -1)
										comps.elementAt(minID).addPixel(x, y);
						}
				}

				MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(binImg
				    .getSizeX(), binImg.getSizeY(), 1, 1, 1, MTBImageType.MTB_BYTE);
				for (int i = 0; i < comps.size(); ++i) {
						MTBRegion2D r = comps.elementAt(i);
						Vector<Point2D.Double> points = r.getPoints();
						for (int j = 0; j < points.size(); ++j) {
								Point2D.Double p = points.elementAt(j);
								resultImage.putValueInt((int) Math.round(p.getX()), (int) Math.round(p
								    .getY()), i + 1);
						}
				}
				return resultImage;
		}
}
