/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.operator;

import de.unihalle.informatik.MiToBo.exceptions.*;

import java.util.*;
import java.io.*;

import org.apache.xmlbeans.*;

import org.graphdrawing.graphml.xmlns.*;

import de.unihalle.informatik.MiToBo.exceptions.MTBException;


/** This is the abstract class from which all data classes to be passed into or out of
  * opertors have to be derived. The most prominent example are images.
  * <p>
  * Each MTBData object has a history which is read from file or
  * may be empty if the data are created from scratch.
  * <p>
  * In addition to the data itself MTBData holds properties like image type,
  * resolution or specification of acquisition devices.
  * There is one specific property with key "location" which is expected
  * to contain a file name or in general URI to the persistant storage
  * location. Applicable only if this data object was read from persistant storage
  * and not created from scratch.
  * Properties may change during the lifetime of a MTBData object.
  * <p>
  * To each MTBData object there is a unique data port associated which
  * represents the data as input source for operators.
  * This data port will in general change during processing as a data object
  * is passed into and out of operators.
  */

abstract public class MTBData {
	private boolean verbose = false;

	/** extension of mitobo processing history file
	 */
	public final String MPH_EXTENSION = new String( "mph");

	/** Properties of the data object
      */
	private Hashtable<String, String> properties;

	/** The processing history of this data object
	 *  stemming from a graphML file. 
	 *  Maybe null if none could be read from file or none was created.
	 * Note: this does NOT and will never comprise the processing within the current
	 * process, rather an ancient processing history read from file or other persistant
	 * storage media.
	 */
	private GraphmlType graphmlHistory;

	/** The processing history of this data object
	 *  stemming from a general, non-graphML XML-file. 
	 *  Maybe null if graphMLHistory exists or non could be read from file
	 *  or none was created.
	 * Note: this does NOT and will never comprise the processing within the current
	 * process, rather an ancient processing history read from file or other persistant
	 * storage media.
	 */
	private XmlObject xmlHistory;

	/** This is the data port representing this data object.
	 * This data port will in general change during processing as a data object
	 * is passed into and out of operators.
	 */
	private MTBPort dataport;

	/** Create a data object.
	 * Initializes the member varaiables as appropriate.
	 */
	public MTBData() {
		this.properties = new Hashtable<String, String>();
		graphmlHistory = null;
		xmlHistory = null;
		dataport = new MTBDataPort( this);
	}


	/** Read and set the processing history from file. If no graphML processing history
	 *  could be read from, an arbitrary XML tree is read. If also this fails,
	 *  the data object is left without any processing history.
	 *  As a result of a call to readHistory either graphmlHistory
	 *  is set, or xmlHistory is set, or none is set.
	 *  The extension <code>mph</code> is used for the history filename and replaces the extension in <code>filename</code>
	 *  or is appended to <code>filename</code> if <code>filename</code> does not contain a extension.
	 * <p>
	 * Additionally the location property is set from filename.
	 *
	 * @param	filename	filename of processing history (XML file)
	 */
	public void readHistory(String filename) {
		GraphmlDocument graphmlDoc = null;

		setLocation( filename);

		String filenameWOExtension = removeExtension( filename);
		try {
			File file = new File( filenameWOExtension  + "." + MPH_EXTENSION);
			graphmlDoc = GraphmlDocument.Factory.parse(file);

			graphmlHistory = graphmlDoc.getGraphml();
			((MTBDataPort)dataport).graphmlHistory = graphmlHistory;

			if ( verbose ) {
				System.out.println( "readHistory::filename " + file.getCanonicalPath());
				GraphmlHelper.printGraphml( graphmlHistory);
				System.out.println( "readHistory::filename DONE");
			}
			
		} catch ( Exception e ){
			// try to read a general XML file instead
			graphmlHistory = null;
			if ( verbose ) {
				System.out.println( "Cannot load graphml from " + filenameWOExtension  + "." + MPH_EXTENSION + ", trying general XML");
			}

			try {
				File file = new File( filename + ".xml");
            	xmlHistory = XmlObject.Factory.parse( file);
			} catch ( Exception e2 ){
				if ( verbose ) 
					System.out.println( "Cannot load xml, giving up to load history");

				xmlHistory = null;
        	}
        }
	}

	/** Write the processing history if any to a graphml file. 
	 *  The processing history is created using the opNodes and their ports
	 *  starting with this data object.
	 *  If no history is avaiable as this MTBData is created internally and not output of an 
	 *  operator, then no history file is written.
	 *  The extension <code>mph</code> is used and replaces the extension in <code>filename</code>
	 *  or is appended to <code>filename</code> if <code>filename</code> does not contain a extension.
	 *  If <code>filename</code> already has extension <code>mph</code> an exception in thrown.
	 *
	 * @param	filename	filename to write the processing history (as graphml/XML file) into
	 */
	public void writeHistory( String filename) 
		throws MTBProcessingDAGException,MTBOperatorException {

		if ( getExtension( filename).equals( MPH_EXTENSION) )
			throw new MTBOperatorException( MTBOperatorException.OperatorExceptionType.ILLEGAL_EXTENSION, filename);

		MTBProcessingDAG pDAG = new MTBProcessingDAG();
		GraphmlDocument graphmlDoc = pDAG.createGraphmlDocument( this);

		// any history to write?
		if ( graphmlDoc == null ) return;

		try {
			String filenameWOExtension = removeExtension( filename);
			BufferedWriter xmlFile =
                new BufferedWriter(new FileWriter( filenameWOExtension + "." + MPH_EXTENSION));

			xmlFile.write( graphmlDoc.toString());
			xmlFile.close();
		} catch ( Exception e ){
			System.err.println( "Exception" + e );
			e.printStackTrace();
		}
	}

	/** Get the port which currently represents this data
	 *
	 * @return current dataport of this data
	 */
	public MTBPort getDataport() { 
		return dataport;
	}

	/** Set the port which currently represents this data
	 *
	 * @param port new current dataport of this data
	 */
	void setDataport( MTBPort port) { 
		dataport = port;
	}


	/** Get the keys of all properties set in this data object
	 *
	 * @return	property keys
	 */
	public Enumeration<String> getPropertyKeys() {
		return this.properties.keys();
	}

	/** Set the property with key to new value o.
	 * Set the property also for associated port, if it is a MTBDataPort.
	 *
	 * @param key	key of property to set
	 * @param o		new value of this property
	 */
	public void setProperty( String key, Object o) {
		this.properties.put( key, o.toString());

		// set property also for associated port, if it is a MTBDataPort
		if ( dataport instanceof MTBDataPort )
			((MTBDataPort)dataport).setProperty( key, o.toString());
	}

	/** Get a property value for the given <code>key</code>.
	 *
	 * @param key	key of property to get
	 */
	public String getProperty( String key) {
		return this.properties.get( key);
	}

	/** Get a property value for the given <code>key</code>.
	 *
	 * @param key	key of property to get
	 */
	public Hashtable<String, String>  cloneProperties() {
		return (Hashtable<String, String>)(this.properties.clone());
	}

	/** Set the location property, i.e. property with key location
	 *
	 * @param location		new value of location property
	 */
	public void setLocation( String location) {
		this.properties.put( "location", location);
        try { 
			((MTBDataPort)dataport).setProperty( "location", location);
		} catch ( Exception e ){
		}
	}

	/** Get the location property, i.e. property with key location
	 */
	public String getLocation() {
		return this.properties.get( "location");
	}

	/** Get the processing history of this data object as read from a graphml file
	 */
	public GraphmlType getGraphmlHistory() {
		return this.graphmlHistory;
	}

	/** Get the processing history of this data object as read from a generic XML file,
	 *  which is not a graphml file
	 */
	public XmlObject getXmlHistory() {
		return this.xmlHistory;
	}

	/** Print this data object to standard out
	 */
	public void print() {
		System.out.println( "MTBData" + this);
		Enumeration<String> keys = properties.keys();

        while ( keys.hasMoreElements() ) {
			String key = keys.nextElement();
			System.out.println( "    " + key + " --> " + getProperty( key));
        }
    }

	/** Remove extension from filename if any
	 * @param	str		filename
	 * @return	filename without extension
	 */
	public static String removeExtension(String str) {
		String separator = System.getProperty("file.separator");
		String filename;
		String path;   // str before last file.separator

		// Remove the path upto the filename.
		int lastSeparatorIndex = str.lastIndexOf(separator);
		if (lastSeparatorIndex == -1) {
			filename = str;
			path = new String("");
		} else {
			filename = str.substring(lastSeparatorIndex + 1);
			path = str.substring( 0, lastSeparatorIndex+1);
		}
	
		// Remove the extension.
		int extensionIndex = filename.lastIndexOf(".");
		if (extensionIndex == -1)
			return path + filename;
	
		return path + filename.substring(0, extensionIndex);
	}

	/** Get extension from filename if any
	 * @param	str		filename
	 * @return	extension of empty string
	 */
	public static String getExtension(String str) {
		String separator = System.getProperty("file.separator");
		String filename;
		String path;   // str before last file.separator

		// Remove the path upto the filename.
		int lastSeparatorIndex = str.lastIndexOf(separator);
		if (lastSeparatorIndex == -1) {
			filename = str;
			path = new String("");
		} else {
			filename = str.substring(lastSeparatorIndex + 1);
			path = str.substring( 0, lastSeparatorIndex+1);
		}
	
		// Remove the extension.
		int extensionIndex = filename.lastIndexOf(".");
		if (extensionIndex == -1)
			return new String("");
		else
			return filename.substring(extensionIndex+1);
	}

	/** Remove all leading pathname components from a filename
	 * @param	str		filename
	 * @return	tail fo the filename
	 */
	public static String removeLeadingDirectories(String str) {
		String separator = System.getProperty("file.separator");
		String filename;
		String path;   // str before last file.separator

		// Remove the path upto the filename.
		int lastSeparatorIndex = str.lastIndexOf(separator);
		if (lastSeparatorIndex == -1) {
			filename = str;
			path = new String("");
		} else {
			filename = str.substring(lastSeparatorIndex + 1);
			path = str.substring( 0, lastSeparatorIndex+1);
		}
	
		return filename;
	}

}
