/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.operator;

import de.unihalle.informatik.MiToBo.exceptions.*;

import ij.IJ;
import ij.ImagePlus;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;

import de.unihalle.informatik.MiToBo.datatypes.images.*;

/**
 * Base class for MiToBo plugin classes. 
 * <p>
 * This class is supposed to alleviate the development of self-documenting plugins in MiToBo. 
 * If one would like to implement a plugin that
 * <ul>
 * <li> gets a single image as input
 * <li> has no specific parameters to document
 * <li> returns only a single output image
 * </ul>
 * then you can simply extend this class. It handles the basic stuff that one would usually
 * do in a MiToBo self-documenting plugin:
 * <ul>
 * <li> copy input image and image processor to local variables 'inputImage' and 'inputProc'
 * <li> initialize a dummy parameter object for the plugin
 * <li> get the meta data from the input image
 * <li> copy the meta object, add it to the dummy parameters and pass both to the 'runOp' method
 * <li> get the result image and attach the meta data object to it
 * <li> display the result image
 * </ul>
 * The only thing that remains to do for you is to implement the 
 * routine {@link de.unihalle.informatik.MiToBo.meta.MTBOpBase#operate(MTBOpParams)}.
 * <p>
 * Please note, there are a few things that you have to take into account:
 * <ul>
 * <li> {@link de.unihalle.informatik.MiToBo.meta.MTBOpBasePlugin#run(ImageProcessor)} in this 
 * 		class assumes the result image to be stored in the parameter object with identifier 
 *      'mtb_resultimage'; if this is not the case, no result image is shown
 * <li> the default setup-routine returns {@link ij.plugin.filter.PlugInFilter#DOES_ALL};
 * 	    if your plugin has limitations you have to overwrite the method and return the correct value
 * </ul>
 * 
 * @author moeller
 */
public abstract class MTBOperatorPlugInFilter 
	extends MTBOperator implements PlugInFilter {

	protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		@SuppressWarnings("hiding")
	  MTBOpArgumentDescriptor[] parameterDescriptors = null;
	
		@SuppressWarnings("hiding")
	  MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
																		"Input image", true, null)
			};
	
		@SuppressWarnings("hiding")
	  MTBOpArgumentDescriptor[] outputDescriptors = null;
		@SuppressWarnings("hiding")
	  MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
			addDescriptors( parameterDescriptors, inputDescriptors,
							 outputDescriptors, supplementalDescriptors);
	}
	
	/**
	 * Input image (ImageJ format).
	 */
	protected ImagePlus inputImage;
	
	/**
	 * Image processor of input image.
	 */
	protected ImageProcessor inputProc;

	/**
	 * Input image converted to MTBImage format.
	 */
	protected MTBImage mtbInput;
	
	/** 
	 *  Constructor. 
	 */
	public MTBOperatorPlugInFilter() throws MTBOperatorException {
	}

	@Override
	public int setup(String arg, ImagePlus imp) {
		if (imp == null) {
				IJ.showMessage("No image available, open one first!");
				return DOES_ALL;
		}
		this.inputImage= imp;
		this.inputProc= imp.getProcessor();
		this.mtbInput= MTBImage.createMTBImage(imp);
		return DOES_ALL;
	}
}
