/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.operator;

import de.unihalle.informatik.MiToBo.exceptions.*;

import java.util.*;
import java.io.*;

import org.graphdrawing.graphml.xmlns.*;

import org.apache.xmlbeans.*;

public class MTBProcessingDAG {

	/** verbose flag for debugging output
	 */
	private boolean verbose = false;

	/** Character separating graphId prefix from rest
	 */
	public static final char idSeparator = ':';

	//private GraphmlType graphMl;
	private GraphmlHelper gHelper;

	/** This index is the index used to build the prefix for all Ids for elements
	 *  for the next history to add to the current graphml.
	 *  The currently build graph has index 0.
	 */
	private int graphIndex; // in the future this to 

	private Vector<MTBOpNode> allOpNodes; // opNodes found in this processing DAG
	private Vector<MTBOpNode> opNodesClone; // opNodes clone to generate unique Ids 
	private Vector<MTBEdge> allEdges; // edges in processing DAG
	private Vector<MTBDataPort> allDataports; // all data in processing DAG
	private Vector<MTBDataPort> writtenDataports; // data already written to the DAG

	//private Vector<MTBPort> ports; // ports in this processing DAG

	private	MTBData	outputData; // the MTBData object for this processing history 

	//private ObjectFactory of;
		
	/** Set the verbose state 
	 * @param   verbose     New verbose state
	 */
	public void setVerbose( boolean verbose) {
		this.verbose = verbose;
	}

	/** Return the verbose state 
	 * @return	Current verbose state
	 */
	public boolean getVerbose() {
		return this.verbose;
	}

	/** Reconstruct processing DAG for data 
	 * @param	data	MTBData for which the processing history to be written.
	 */
	private void createDAG( MTBData data) throws MTBProcessingDAGException {
		graphIndex = 1;
		outputData = data;
		allOpNodes = new Vector<MTBOpNode>();
		allEdges = new Vector<MTBEdge>();
		allDataports = new Vector<MTBDataPort>();
		writtenDataports = new Vector<MTBDataPort>();

		if ( data.getDataport() == null ) {
			System.err.println( "MTBProcessingDAG::createDAG data has no output port");
			throw new MTBProcessingDAGException(MTBProcessingDAGException.DAGExceptionType.INTERNAL_TRACING_ERROR,null);
		}

		if (  data.getDataport() instanceof MTBDataPort ) {
			allDataports.add( (MTBDataPort)data.getDataport());
			return;
		}

		visitOpNodeBottomUp( ((MTBOpNodePort)(data.getDataport())).getOpNode(), 0);
		
		// clone of opNode vector used to generate unique Ids for opnodes
		opNodesClone = (Vector<MTBOpNode>)(allOpNodes.clone());

		if ( verbose ) {
			System.out.println( "MTBProcessingDAG::createDAG");
			System.out.println( ">>>>>>>>>>>>> opNodes found");
			Iterator<MTBOpNode> oItr = allOpNodes.iterator();
			while ( oItr.hasNext() ) {
				oItr.next().print();
			}

			System.out.println( ">>>>>>>>>>>>> edges found");
			Iterator<MTBEdge> eItr = allEdges.iterator();
			while ( eItr.hasNext() ) {
				eItr.next().print();
			}

			System.out.println( ">>>>>>>>>>>>> data found");
			Iterator<MTBDataPort> dItr = allDataports.iterator();
			while ( dItr.hasNext() ) {
				System.out.println( dItr.next());
			}
		}
	}

	//----------------------------
	private MTBOpNode tracebackInputPorts( MTBInputPort[] inputPorts,  MTBOpNode parentOpNode, 
			Vector<MTBOpNode> sibblings, Vector<MTBDataPort> dataports, int depth) {

		MTBPort origin;

		for ( int i=0;  i < inputPorts.length ; i++ ) {
			MTBInputPort inputPort = inputPorts[i];

			origin = inputPort.getOrigin();
				
			if ( origin != null ) {		
				
				allEdges.add( new MTBEdge( origin, inputPort));
	
				if ( origin instanceof MTBInputPort ) {
					if ( parentOpNode == null ) {
						parentOpNode = ((MTBInputPort)origin).getOpNode();
					} else {
						if ( ((MTBInputPort)origin).getOpNode() != parentOpNode )
							System.err.println( "MTBProcessingDAG::tracebackInputPorts ERROR; found different parentNodes");
					}
	
				} else if ( origin instanceof MTBOutputPort ) {
					MTBOpNode opNode = ((MTBOutputPort)origin).getOpNode();
					if ( ! allOpNodes.contains( opNode) ) {
						sibblings.add( opNode);
						visitOpNodeTopDown( opNode, depth);
						MTBOpNode newParentOpNode = tracebackInputPorts( opNode.getInputPorts(), parentOpNode, sibblings, dataports, depth);
						if ( parentOpNode == null ) {
							parentOpNode = newParentOpNode;
						} else {
							if ( newParentOpNode != parentOpNode )
								System.err.println( "MTBProcessingDAG::tracebackInputPorts ERROR; found different parentNodes");
						}
					}
				} else if ( origin instanceof MTBDataPort ) {
					// potentail check; should only be found once
					if ( ! allDataports.contains( origin) ) {
						allDataports.add( (MTBDataPort)origin);
					}
					dataports.add( (MTBDataPort)origin);
				} else { 
				}
			}
		}
	
		return parentOpNode;
	}

	//----------------------------
	//private void tracebackOpNodePort( MTBOpNodePort outputPort, int depth) {
	private void visitOpNodeBottomUp( MTBOpNode opNode, int depth) {
		// opNodes, sibblings, and dataports are either in the same parentOpNode, which is still
		// to be discovered, or in the top level of the graph
		Vector<MTBOpNode> sibblings = new Vector<MTBOpNode>();
		Vector<MTBDataPort> dataports = new Vector<MTBDataPort>();
	
		MTBOpNode parentOpNode = null;
	
		if ( ! allOpNodes.contains( opNode) ) {
			visitOpNodeTopDown( opNode, depth);  
			MTBOpNode newParentOpNode = tracebackInputPorts( opNode.getInputPorts(), parentOpNode, sibblings, dataports, depth);
			if ( parentOpNode == null ) {
				parentOpNode = newParentOpNode;
			} else  {
				if ( newParentOpNode != parentOpNode )
					System.err.println( "MTBProcessingDAG::tracebackOutputPort ERROR; found different parentOpNodes");
			}
		}

		if ( parentOpNode != null )  {
			register( parentOpNode, opNode);
			for ( int s=0 ; s <  sibblings.size() ; s++ ) {
				MTBOpNode sibbling = sibblings.elementAt( s);
				register( parentOpNode, sibbling);
			}

			// merke: wir muessen am ende die nicht registrierten dataports auch schreiben
			// die sollten dann alle auf der niedrigsten depth sein, also auf hoechster Graphebene
			// Vielleicht sollten wir den dataports auch eine tiefe geben??
			for ( int d=0 ; d <  dataports.size() ; d++ ) {
				MTBDataPort dataport = dataports.elementAt( d);
				register( parentOpNode, dataport);
			}

			if ( ! allOpNodes.contains( parentOpNode) ) {
				visitOpNodeBottomUp( parentOpNode, depth-1);
			}
		}
	}

	//-------------------
	// we call visitOpNodeTopDown not twice for the same opNode
	private void visitOpNodeTopDown( MTBOpNode opNode, int depth) {
		allOpNodes.add( opNode);
		opNode.setDepth( depth);

		LinkedList<MTBOpNode> childrenToVisit = new LinkedList<MTBOpNode>( opNode.getDirectlyRegisteredChildern());

		MTBOutputPort[] outputPorts = opNode.getOutputPorts();
		for ( int i=0 ; i < outputPorts.length ; i++ )
			handlePort( opNode, outputPorts[i], childrenToVisit);

		while( childrenToVisit.size() > 0 ) {
			MTBOpNode childOpNode = childrenToVisit.pop();

			// potentially check, if childOpNode.getParent() is nonnull but different from opNode ??
			// --> general check in method register()
			register( opNode, childOpNode);
			if ( ! allOpNodes.contains( childOpNode) ) {
				visitOpNodeTopDown( childOpNode, depth+1);

				MTBInputPort[] inputPorts = childOpNode.getInputPorts();
				for ( int i=0 ; i < inputPorts.length ; i++ )
					handlePort( opNode, inputPorts[i], childrenToVisit);
			}
		}
	}

	
	// we visit each opNode not more than once
	//private void visitOpNodeBottomUp( MTBOpNode opNode, int depth) {
		//visitOpNodeTopDown( opNode, depth);
//
		//MTBInputPort[] inputPorts = opNode.getInputPorts();
		//for ( int i=0 ; i < inputPorts.length ; i++ ) {
			//MTBPort origin = inputPorts[i].getOrigin();
			//if ( origin != null )  {
				//allEdges.add( new MTBEdge( origin, inputPorts[i]));
//	
				//if ( origin instanceof MTBInputPort ) {
					//MTBOpNode parentOpNode = ((MTBInputPort)origin).getOpNode();
					//// check something??
					//register( parentOpNode, opNode);
					//if ( ! allOpNodes.contains( parentOpNode) ) {
						//visitOpNodeBottomUp( parentOpNode, depth-1);
					//}
				//} else if ( origin instanceof MTBOutputPort ) {
					//tracebackOpNodePort( ((MTBOutputPort)origin), depth);
				//} else { // DataPort
					//// potential check: dataport should be discovered only once
					//if ( ! allDataports.contains( origin) ) {
						//register( opNode, (MTBDataPort)origin);
						//allDataports.add( (MTBDataPort)origin);
					//}
				//}
			//}
		//}
	//}

	//
	private void register( MTBOpNode parentOpNode, MTBOpNode opNode) {
		if ( opNode.getParent() == null ) {
			opNode.setParent( parentOpNode);
			parentOpNode.addChild( opNode);
		} else {
			if ( opNode.getParent() != parentOpNode )
				System.out.println( "MTBProcessingDAG::register register parent for " + opNode + " again with different parent");
		}
	}
	

    /**
	 */
	private void register( MTBOpNode opNode, MTBDataPort dataport ) {
		if ( ! opNode.getIncludedData().contains( dataport ) )
			opNode.addData( dataport);
	}

	/**
	 */
	private void handlePort( MTBOpNode opNode, MTBPort port, LinkedList<MTBOpNode> childrenToVisit) {
		MTBPort origin = port.getOrigin();
		if ( origin != null ) {
			allEdges.add( new MTBEdge( origin, port));

			if ( origin instanceof MTBInputPort ) {
				// potential check: origin.getOpNode == opNode
				// nothing to be done here, handled outside
			} else if ( origin instanceof MTBOutputPort ) {
				MTBOpNode childOpNode = ((MTBOutputPort)origin).getOpNode();
				// potenital check: if childOpNode has already a depth, childOpNode.getDepth = depth+1
				childrenToVisit.add( childOpNode);
			} else if ( origin instanceof MTBDataPort ) {
				// potential check: dataport should be discovered only once
				if ( ! allDataports.contains( origin) ) {
					register( opNode, (MTBDataPort)origin);
					allDataports.add( (MTBDataPort)origin);
				}
			} else { // DataPort
			}
		}
	}

	// ===================== convert to graphml structure
	/** Create the processing history for <cod>data</code> and convert to graphml object.
	 *
	 * @param	data	MTBData for which the processing history to be written.
	 * @return Processing history as graphml object or null if no history is available.
	 */
	public GraphmlDocument createGraphmlDocument( MTBData data) 
		throws MTBProcessingDAGException {

		if ( verbose ) 
			System.out.println( "MTBProcessingDAG::createGraphmlDocument START for " + data);

		createDAG( data);

		GraphmlDocument graphmlDoc = GraphmlDocument.Factory.newInstance();
		gHelper = new GraphmlHelper( );

		// setup the graphml
		GraphmlType graphMl = graphmlDoc.addNewGraphml();
		graphMl.setDesc( "G");

		// keys
		addKey( graphMl, "mtbName", KeyForType.NODE, "mtbName", KeyTypeType.STRING);
		addKey( graphMl, "mtbNodeType", KeyForType.NODE, "mtbNodeType", KeyTypeType.STRING);
		addKey( graphMl, "mtbExplanation", KeyForType.NODE, "mtbExplanation", KeyTypeType.STRING);
		addKey( graphMl, "mtbClassname", KeyForType.NODE, "mtbClassname", KeyTypeType.STRING);
		addKey( graphMl, "mtbProps", KeyForType.NODE, "mtbProperties", KeyTypeType.STRING);
		addKey( graphMl, "mtbParas", KeyForType.NODE, "mtbParameteres", KeyTypeType.STRING);
		addKey( graphMl, "mtbParasXML", KeyForType.NODE, "mtbParameteresAsXml", KeyTypeType.STRING);
		addKey( graphMl, "mtbOutput", KeyForType.GRAPH, "mtbOutput", KeyTypeType.STRING);
		addKey( graphMl, "mtbXml", KeyForType.GRAPH, "mtbXmlHistory", null);

		// top level graph
		GraphType graph = graphMl.addNewGraph();
		graph.setEdgedefault( GraphEdgedefaultType.DIRECTED);

		// add output port of MTBData of this processing history as attribute 
		addAttrToGraph( graph, "mtbOutput", getPortId( outputData.getDataport()));


		// hierarchy of opNodes
		int minDepth = 1;
		for ( int i=0 ; i < allOpNodes.size() ; i++ ) {
			if ( allOpNodes.elementAt(i).getDepth() < minDepth ) 
				minDepth = allOpNodes.elementAt(i).getDepth();
		}

		while ( allOpNodes.size() > 0 ) {
			boolean found = false;
			for ( int i=0 ; i < allOpNodes.size() ; i++ ) {
				if ( allOpNodes.elementAt(i).getDepth() == minDepth ) {
					outputOpNode( allOpNodes.elementAt(i), graph);
					found = true;
					// we restart, as we have modified opNodes during outputOpNode
					break;
				}
			}

			// we did not find any node at depth 1
			if ( ! found ) break;
		}

		if ( allOpNodes.size() > 0 ) {
			System.err.println( "MTBProcessingDAG::createGraphmlDocument WARNING: leftover orphan opNodes");
			for ( int i=0 ; i < allOpNodes.size() ; i++ ) {
				allOpNodes.elementAt(i).print();
			}
		}

		// MTBdata which have not already been written
		Iterator<MTBDataPort> dItr = allDataports.iterator();
		while ( dItr.hasNext() ) {
			MTBDataPort d = dItr.next();
			if ( ! writtenDataports.contains( d) ) {
				if ( verbose ) 
					System.out.println( "MTBProcessingDAG::createGraphmlDocument remaining MTBData " + d);
				addDataPortToGraphml( graph, d);
			}
		}

		// edges
		int idx = 0;
		Iterator<MTBEdge> eItr = allEdges.iterator();
		while ( eItr.hasNext() ) {
			MTBEdge e = eItr.next();
			int n = allEdges.indexOf( e);
			
			addEdge( graph, e.getSourcePort(), e.getTargetPort(), idx);
			idx++;
		}

		return graphmlDoc;
	}

	// ===================== Helper methods for Write DAG

	/** Output opNode and recursively its children
	 */
	private void outputOpNode( MTBOpNode opNode, GraphType graph) 
		throws MTBProcessingDAGException {

		if ( verbose ) 
			System.out.println( "MTBProcessingDAG::outputOpNode start output for " + opNode);

		allOpNodes.remove( opNode);

		NodeType node = addOpNodeToGraph( graph, opNode);
        GraphType nestedG = node.addNewGraph();
        nestedG.setEdgedefault( GraphEdgedefaultType.DIRECTED);
		
		// ports
		MTBInputPort[] inputPorts = opNode.getInputPorts();
		for ( int p=0 ; p < inputPorts.length ; p++ ) {
			// port as graphml nodes
			addOpNodePortToGraphml( nestedG, inputPorts[p]);
		}

		MTBOutputPort[] outputPorts = opNode.getOutputPorts();
		for ( int p=0 ; p < outputPorts.length ; p++ ) {
			// port as graphml nodes
			addOpNodePortToGraphml( nestedG, outputPorts[p]);
		}	

		// children found during traceback
		Vector<MTBOpNode> children = opNode.getChildren();
		Iterator<MTBOpNode> oItr = children.iterator();
		while (oItr.hasNext()) {
			MTBOpNode child = oItr.next();
			if ( allOpNodes.contains( child) ) {
				allOpNodes.remove( child);
				outputOpNode( child, nestedG);
			}
		}

		// included data
		Iterator<MTBDataPort> dItr = opNode.getIncludedData().iterator();
		while ( dItr.hasNext() ) {
			MTBDataPort d = dItr.next();

			addDataPortToGraphml( nestedG, d);

			// delete from allDataports, to avoid multiple graph nodes for this MTBData
			writtenDataports.add( d);
		}
	}

	private void addEdge( GraphType graph, MTBPort sourcePort, MTBPort targetPort, int idx) 
		throws MTBProcessingDAGException {
		EdgeType edge = graph.addNewEdge();
		edge.setId( getGraphName(0) + idSeparator + "Edge" + idx);

		// ports as graphml ports
		edge.setSource( getPortId( sourcePort));
		edge.setTarget( getPortId( targetPort));

		//addAttrToEdge( edge, "arrow", "Target");
		edge.setDirected( true);
	}

	// =========================== histories ==========================
	/** Add the complete history from a graphml file of a MTBData to a graph object.
	 *  Add all nodes and edges (besides the toplevel graph) from the
	 *  history to the graph object at the toplevel of the graph and
	 *  connect the output port to the data port
	 */
	private void addGraphmlHistory( GraphType graph, MTBDataPort dataport) 
		throws MTBProcessingDAGException {
		if ( verbose ) 
			System.out.println( "MTBProcessingDAG::addGraphmlHistory");

		//MTBData data = dataport.getData();

		if ( dataport.getGraphmlHistory() == null ) return;
		
		GraphmlType history = dataport.getGraphmlHistory();

		GraphType hGraph = GraphmlHelper.getToplevelGraph( history);
		if ( hGraph == null ) return;

		// rename Ids in history graphml
		GraphmlHelper.renameGraphIds( hGraph, getGraphName( graphIndex));
		graphIndex++;

		int oldNumNodes = graph.sizeOfNodeArray();
		for ( int n = 0 ; n < hGraph.sizeOfNodeArray() ; n++ ) {
			NodeType node = hGraph.getNodeArray( n);
			if ( verbose ) 
				System.out.println( "MTBProcessingDAG::addGraphmlHistory add Node ");// + node);
			graph.insertNewNode( oldNumNodes+n);
			graph.setNodeArray( oldNumNodes+n, node);
		}

		int oldNumEdges = graph.sizeOfEdgeArray();
		for ( int e = 0 ; e < hGraph.sizeOfEdgeArray() ; e++ ) {
			EdgeType edge = hGraph.getEdgeArray( e);
			if ( verbose ) 
				System.out.println( "MTBProcessingDAG::addGraphmlHistory add Edge ");// + edge);
			graph.insertNewEdge( oldNumEdges+e);
			graph.setEdgeArray( oldNumEdges+e, edge);
		}

		int oldNumDatas = graph.sizeOfDataArray();
		for ( int d = 0 ; d < hGraph.sizeOfDataArray() ; d++ ) {
			DataType gdata = hGraph.getDataArray( d);

			// add extra edge to connect graphs
			if ( gdata.getKey().equals("mtbOutput") ) {
				EdgeType edge = graph.addNewEdge();
				edge.setSource( gdata.newCursor().getTextValue());
				edge.setTarget( getPortId( dataport));
				edge.setDirected(true);
				
				if ( verbose ) 
					System.out.println( "MTBProcessingDAG::addGraphmlHistory add Edge for Output " + 
						gdata.newCursor().getTextValue() + " --> " + getPortId( dataport));
			}
		}
	}

	/** Add the complete history of a MTBData from a generic XML file to a graphml node.
	 *  The complete xml tree is put in place without any processing.
	 */
	private void addXmlHistory( NodeType node, XmlObject xmlHistory) {
		
		if ( verbose ) 
			System.out.println( "MTBProcessingDAG::addXmlHistory");

		XmlCursor hCursor = xmlHistory.newCursor();
		hCursor.toStartDoc();
		hCursor.toNextToken();


		DataType data = node.addNewData();
        data.setKey( "mtbXml");
		XmlCursor dCursor = data.addNewXmlHistory().newCursor();
		dCursor.toEndToken();
		hCursor.moveXml( dCursor);

	}

	/** Add the parameter hash to to a graphml node.
	 */
	private void addParameterHash( NodeType node, MTBOpNode opNode) {
		
		if ( verbose ) 
			System.out.println( "MTBProcessingDAG::addParameterHash");

		XmlCursor hCursor = opNode.getParameterHashAsXml().newCursor();
		hCursor.toStartDoc();
		hCursor.toNextToken();


		DataType data = node.addNewData();
        data.setKey( "mtbParasXML");
		XmlCursor dCursor = data.addNewParameterWrapper().newCursor();
		dCursor.toEndToken();
		hCursor.copyXml( dCursor);

	}

	//  ======================== graphml Helper ================================
	/** add a key definition to graphml
	 */
	public void addKey( GraphmlType graphml, String id, KeyForType.Enum domain, String name, KeyTypeType.Enum keyType) {
		KeyType key = graphml.addNewKey();
		key.setId( id);
		key.setFor( domain);
		key.setAttrName( name);
		key.setAttrType( keyType);
	}

	/** add an opNode to the graphml within the graph
	 */
	public NodeType addOpNodeToGraph( GraphType graph, MTBOpNode opNode) {
		NodeType node = graph.addNewNode();
		node.setId( getOpNodeId( opNode));

		addAttrToNode( node, "mtbName", opNode.getName());
		addAttrToNode( node, "mtbNodeType", "OpNode");

		// parameteres
		DataType data = node.addNewData();
		data.setKey( "mtbParas");
		PropertyListType props = data.addNewProperties();

		// commented for the moment, as it did crash
		//addParameterHash( node, opNode);

		Enumeration<String> keys = opNode.getParameterKeys();
		while ( keys.hasMoreElements() ) {
			String key = keys.nextElement();
			String value = opNode.getParameter( key);

			PropertyType prop = props.addNewProperty();
			prop.setKey( key);
			prop.setValue( value);
		}

		// properties
		data = node.addNewData();
		data.setKey( "mtbProps");
		props = data.addNewProperties();

		PropertyType prop = props.addNewProperty();
		prop.setKey( "classname"); 
		prop.setValue( opNode.getOperatorClass().getSimpleName());

		if ( opNode.getOperatorClass().getPackage() != null ) {
			prop = props.addNewProperty();
			prop.setKey( "package"); 
			prop.setValue( opNode.getOperatorClass().getPackage().getName());
		}

		prop = props.addNewProperty();
		prop.setKey( "version"); 
		prop.setValue( opNode.getVersion());

		return node;
	}

	/** Add a MTBData object to the graphML graph
	 */

	public NodeType addDataPortToGraphml( GraphType graph, MTBDataPort dataport) 
		throws MTBProcessingDAGException {
		//MTBData mtbData = dataport.getData();
		NodeType node = graph.addNewNode();
		node.setId( getDataPortId( dataport));

		String loc;
		if ( (loc = dataport.getLocation()) != null )
			addAttrToNode( node, "mtbName", MTBData.removeLeadingDirectories( loc));
		else
			addAttrToNode( node, "mtbName", "Internally created");

		if ( dataport.getGraphmlHistory() != null ) {
			addAttrToNode( node, "mtbNodeType", "DataPortFromFile");
			addGraphmlHistory( graph, dataport);
		} else if ( dataport.getXmlHistory() != null ) {
			addAttrToNode( node, "mtbNodeType", "DataPortFromFile");
			addXmlHistory( node, dataport.getXmlHistory());
		} else {
			addAttrToNode( node, "mtbNodeType", "DataPort");
		}

		DataType data = node.addNewData();
		data.setKey( "mtbProps");
		PropertyListType props = data.addNewProperties();
		Enumeration<String> keys = dataport.getPropertyKeys();
		while ( keys.hasMoreElements() ) {
			String key = keys.nextElement();
			String value = dataport.getProperty( key);
	
			PropertyType prop = props.addNewProperty();
			prop.setKey( key);
			prop.setValue( value);
		}

		return node;
	}

	/** Add a opNode port object to the graphML graph
	 */

	public NodeType addOpNodePortToGraphml( GraphType graph, MTBOpNodePort port) 
		throws MTBProcessingDAGException {
		NodeType node = graph.addNewNode();

		node.setId( getPortId( port));

		MTBOpNode opNode = port.getOpNode();
		String descriptorName = port.getDescriptorName();

		try {
			if ( port instanceof MTBOutputPort ) {
				addAttrToNode( node, "mtbName", descriptorName);
				addAttrToNode( node, "mtbNodeType", "OutputPort");
				addAttrToNode( node, "mtbExplanation", ((MTBOpNodePort)port).getExplanation());
				addAttrToNode( node, "mtbClassname", ((MTBOpNodePort)port).getClassname());
	
				MTBOutputPort outputPort = (MTBOutputPort)port;
				// properties
				DataType data = node.addNewData();
				data.setKey( "mtbProps");
				PropertyListType props = data.addNewProperties();
	
				Enumeration<String> keys = outputPort.getPropertyKeys();
				while ( keys.hasMoreElements() ) {
					String key = keys.nextElement();
					String value = outputPort.getProperty( key);
	
					PropertyType prop = props.addNewProperty();
					prop.setKey( key);
					prop.setValue( value);
				}
			} else { // InputPort
				addAttrToNode( node, "mtbName", descriptorName);
				addAttrToNode( node, "mtbNodeType", "InputPort");
				addAttrToNode( node, "mtbExplanation", ((MTBOpNodePort)port).getExplanation());
				addAttrToNode( node, "mtbClassname", ((MTBOpNodePort)port).getClassname());
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		return node;
	}

	// ======= data attributes ===========================
	public void addAttrToGraphml( GraphmlType graphml, String key, String value) {
		DataType data = graphml.addNewData();

		data.setKey( key);
		data.set( XmlString.Factory.newValue( value));
	}

	public void addAttrToGraph( GraphType graph, String key, String value) {
		DataType data = graph.addNewData();

		data.setKey( key);
		data.set( XmlString.Factory.newValue( value));
	}

	public void addAttrToNode( NodeType node, String key, String value) {
		DataType data = node.addNewData();

		data.setKey( key);
		data.set( XmlString.Factory.newValue( value));
	}

	public void addAttrToEdge( EdgeType edge, String key, String value) {
		DataType data = edge.addNewData();

		data.setKey( key);
		data.set( XmlString.Factory.newValue( value));
	}

	public void addAttrToPort( PortType port, String key, String value) {
		DataType data = port.addNewData();

		data.setKey( key);
		data.set( XmlString.Factory.newValue( value));
	}

	//  ======================== IDs ================================
	/** Remove the graphIP including idSeparator at the beginning of an id 
	 */
	public static String removeGraphId( String id) {
		return id.substring( id.indexOf(idSeparator) +1);
	}

	private String getGraphName( int idx) {
		return new String( "MTBDAG" + idx);
	}

	private String getOpNodeId( MTBOpNode opNode) {
		return getGraphName( 0) + idSeparator + "OpNode" + opNodesClone.indexOf( opNode);
	}

	private String getDataPortId( MTBDataPort dataport) {
		return getGraphName( 0) + idSeparator + "DataPort" + allDataports.indexOf( dataport);
	}

	private String getPortId( MTBPort p) throws MTBProcessingDAGException {
		if ( p instanceof MTBInputPort ) {
			MTBOpNodePort op = (MTBOpNodePort)p;
			return getGraphName( 0) + idSeparator + "InputPort" + op.getPortIndex() + "-of-" + 
					removeGraphId( getOpNodeId( op.getOpNode()));
		} else if ( p instanceof MTBOutputPort ) {
			MTBOpNodePort op = (MTBOpNodePort)p;
			return getGraphName( 0) + idSeparator + "OutputPort" + op.getPortIndex() + "-of-" + 
					removeGraphId( getOpNodeId( op.getOpNode()));
		} else if ( p instanceof MTBDataPort ) {
			MTBDataPort dp = (MTBDataPort)p;
			return getDataPortId( dp);
		} else {
			System.err.println( "MTBProcessingDAG::getPortId PANIC unkwon Porttype");
			throw new MTBProcessingDAGException(MTBProcessingDAGException.DAGExceptionType.INTERNAL_TRACING_ERROR,null);
		}

	}

}
