/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.regions.labeling;

import java.util.*;
import java.lang.Integer;

import ij.*;
import ij.process.*;
import ij.gui.*;

import de.unihalle.informatik.MiToBo.datatypes.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageInt;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion2DSet;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion2DSet.DrawType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.operator.*;

/**
 *  Sequential component labeling technique for 2D images.
 * Foreground pixels are assumed to have a value > 0,
 * all pixels with value <= 0 are assumed to be background.
 * 
 * @author gress
 *
 */
public class LabelComponentsSequential extends MTBOperator {

	private ImagePlus m_labelImage;
	private ImageProcessor m_labelIP;
	
	private MTBImageInt m_labelImg;

	private int m_width;
	private int m_height;
	
  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

	    MTBOpArgumentDescriptor[] parameterDescriptors = new MTBOpArgumentDescriptor[] {
	        new MTBOpArgumentDescriptor( "DiagonalNeighbors", java.lang.Boolean.class, "true for 8-neighborhood, false for 4-neighborhood", true, null)
	   } ;
	
		MTBOpArgumentDescriptor[] inputDescriptors = new MTBOpArgumentDescriptor[] {
	        new MTBOpArgumentDescriptor( "InputImage", MTBImage.class, "Input image", true, null)
		};
		MTBOpArgumentDescriptor[] outputDescriptors = new MTBOpArgumentDescriptor[] {
	        new MTBOpArgumentDescriptor( "ResultingRegions", MTBRegion2DSet.class, "Resulting regions", true, null),
	        new MTBOpArgumentDescriptor( "LabelImage", MTBImage.class, "Label Image", true, null)
		};
	
	    
	        addDescriptors( parameterDescriptors, inputDescriptors,
	                    outputDescriptors, null);
    }

	public LabelComponentsSequential() throws MTBOperatorException {
	}

	public LabelComponentsSequential(MTBImage img, boolean diagonalNeighbors) throws MTBOperatorException {
		this.setInput("InputImage", img);
        this.setParameter("DiagonalNeighbors", diagonalNeighbors);
    }


	/**
	 * Get reference to the current input image.
	 * 
	 * @return	Input image to work on.
	 * @throws MTBOperatorException 
	 */
	protected MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)(this.getInput("InputImage"));
	}

	/**
	 * Get the neighborhood flag.
	 * @return true if 8-neighborhood is used, false if 4-neighborhood is used
	 * @throws MTBOperatorException 
	 */
	protected boolean getDiagonalNeighborsFlag() throws MTBOperatorException {
		return (Boolean)(this.getParameter("DiagonalNeighbors"));
	}

	/**
	 * Attach the vector of resulting regions.
	 * @param regions
	 * @throws MTBOperatorException 
	 */
	protected void setResultingRegions( MTBRegion2DSet regions) throws MTBOperatorException {
		this.setOutput("ResultingRegions", regions);
	}

	/**
	 * Get the vector of resulting regions.
	 * @return
	 * @throws MTBOperatorException 
	 */
	public MTBRegion2DSet getResultingRegions() throws MTBOperatorException {
		return (MTBRegion2DSet)this.getOutput("ResultingRegions");
	}

	/**
	 * Add the label image to the parameter object
	 * @param regions
	 * @throws MTBOperatorException 
	 */
	protected void setLabelImage(MTBImage labelImage) throws MTBOperatorException {
		this.setOutput("LabelImage", labelImage);
	}

	/**
	 * Get label image (of type MTB_INT)
	 * @return
	 * @throws MTBOperatorException 
	 */
	public MTBImage getLabelImage() throws MTBOperatorException {
		return (MTBImage)this.getOutput("LabelImage");
	}

	@Override
	protected void operate() throws MTBOperatorException, MTBProcessingDAGException {
		MTBImage img = getInputImage();
		
		boolean diagNeighbors = getDiagonalNeighborsFlag();

		setResultingRegions(regionSegmentation(img, diagNeighbors));
		
		// draw label image
		DrawRegion2DSet drawer = new DrawRegion2DSet(DrawType.ID_IMAGE, this.getResultingRegions());
		drawer.runOp(this);

		this.setLabelImage(drawer.getResultImage());
	}
	
	/**
	 * @param img				(binary) input image
	 * @param diagonalNeighbors	set true for 8-neighborhood components, set false for 4-neighborhood
	 * @return 					returns a vector of regions	
	 */
	protected MTBRegion2DSet regionSegmentation(MTBImage img, boolean diagonalNeighbors) {
		
		m_width= img.getSizeX();
		m_height= img.getSizeY();
		
		Vector<MTBRegion2D> regions= new Vector<MTBRegion2D>();
		
		// TODO: remove when deprecated methods are removed, where this ImageProcessor is needed
        m_labelIP = null;
		
        
        m_labelImg = (MTBImageInt)MTBImage.createMTBImage(m_width, m_height, 1, 1, 1, MTBImageType.MTB_INT);
        m_labelImg.setTitle(img.getTitle() + " labels");
        
        // western, northwestern, northern, northeastern label value of the actual pixel
        int l_w,l_nw,l_n,l_ne;
        
        // if only 4-neighborhood is regarded
        // set the diagonal neighors to zero to leave them
        // out of consideration in the following algorithm
        l_nw = 0;
        l_ne = 0;
        
        // smallest neigbor-label
        int l_min;
        
        // count of labeled neighbors
        int l_cnt;
        
        // label variable, which is incremented for new labels
		int label = 1;
		
		
		// collision sets for certain labels (vectorindex = label-1)
		Vector< LinkedHashSet<Integer> > collisions = new Vector< LinkedHashSet<Integer> >();
		
		// tmp references to collision sets
		LinkedHashSet<Integer> tmpSet1;
		LinkedHashSet<Integer> tmpSet2;
		// Iterator over the set elements
		Iterator<Integer> setIter;
		Integer labelInt;
		
		// first step: 
		// run line by line through the image
		// and label pixels		
		for (int y = 0; y < m_height; y++) {
			for (int x = 0; x < m_width; x++) {
				
				if (img.getValueDouble(x,y) > 0.0 && m_labelImg.getValueInt(x,y)==0) {
					
					l_cnt = 0;
					l_min = 0;
					
					if (diagonalNeighbors) {
						
						// get relevant neighboring labels (8-neighborhood)
						l_w = getLabel(x-1,y);
						l_nw = getLabel(x-1,y-1);
						l_n = getLabel(x,y-1);
						l_ne = getLabel(x+1,y-1);
					
						// determine smallest label and count labeled neighbors
						if (l_w > 0) {
							l_cnt++;
							l_min = l_w;
						}
						
						if (l_nw > 0) {
							l_cnt++;
							
							if (l_min == 0 || l_nw < l_min) {
								l_min = l_nw;
							}
						}
						
						if (l_n > 0) {
							l_cnt++;
							
							if (l_min == 0 || l_n < l_min) {
								l_min = l_n;
							}
						}
						
						if (l_ne > 0) {
							l_cnt++;
						
							if (l_min == 0 || l_ne < l_min) {
								l_min = l_ne;
							}
						}						
					}
					else {
						
						// get relevant neighboring labels (4-neighborhood)
						l_w = getLabel(x-1,y);
						l_n = getLabel(x,y-1);
					
						// determine smallest label and count labeled neighbors
						if (l_w > 0) {
							l_cnt++;
							l_min = l_w;
						}
						
						if (l_n > 0) {
							l_cnt++;
							
							if (l_min == 0 || l_n < l_min) {
								l_min = l_n;
							}
						}									
					}

					if (l_cnt == 0) {
					// no neighbors of interest are labeled
						
						// give pixel a new label
						m_labelImg.putValueInt(x,y,label);
						
						// create the collision set for that label
						// only containing the new label (-> no collision)
						tmpSet1 = new LinkedHashSet<Integer>();
						tmpSet1.add(label);
						
						// add the new collision set to the vector of collision sets
						collisions.add(tmpSet1);
						
						label++;
					}
					else if (l_cnt == 1) {
					// no label-collision, only one neighbor is labeled, adopt the label
						
						m_labelImg.putValueInt(x, y, l_min);
					}
					else {
					// there might be a label-collision! More than one neighbor is labeled	
						
						
						// assign lowest label to the actual pixel
						m_labelImg.putValueInt(x, y, l_min);
						
						// get the collision set of the label l_min
						tmpSet1 = collisions.get(l_min - 1);
						
						// test for label collisions
						if (l_w > 0 && l_w != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_w - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}
								
						if (l_nw > 0 && l_nw != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_nw - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}						
						
						if (l_n > 0 && l_n != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_n - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}					
																						
						if (l_ne > 0 && l_ne != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_ne - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}											
					}
				}
			}
		}
		
		// run over collision sets and build a simple label mapping array (labelMap):
		// (index of the array - 1) corresponds to the found label
		// the array value corresponds to the value to which the found label is mapped
		int[] labelMapping = new int[label - 1];
		for (int i = 0; i < labelMapping.length; i++) {
			
			// get the collision set for label (i+1)
			tmpSet1 = collisions.get(i);
	
			// set the label mapping to the first label in the set
			setIter = tmpSet1.iterator();
			labelMapping[i] = setIter.next().intValue();		
		}
		
		// clean up
		collisions.clear();
		
		
		// second step: 
		// run line by line through the image and resolve the label collisions,
		// create Region objects and compute its properties
		
		// relates a label (int) to a Region object
		HashMap<Integer,MTBRegion2D> rmap = new HashMap<Integer,MTBRegion2D>();
		Vector<Integer> real_labels = new Vector<Integer>();
		MTBRegion2D act_region;
		int real_label;
		
		for (int y = 0; y < m_height; y++) {
			for (int x = 0; x < m_width; x++) {
				
				label = m_labelImg.getValueInt(x,y);
				
				// if a labeled pixel is found
				if (label != 0) {
					
					real_label = labelMapping[label - 1];
					
					if (rmap.containsKey(real_label)) {
					// update region					
						
						act_region = rmap.get(real_label);
						act_region.addPixel(x, y);
						m_labelImg.putValueInt(x, y, real_label);
					}
					else {
					// new region
						
						act_region = new MTBRegion2D();
						act_region.addPixel(x, y);	
						rmap.put(real_label, act_region);
						
						real_labels.add(real_label);
						m_labelImg.putValueInt(x, y, real_label);
					}				
				}
			}
		}
		
		m_labelImg = null;
		
		// build vector of regions
		for (int i = 0; i < real_labels.size(); i++) {
			act_region = rmap.get(real_labels.get(i));
			act_region.setID(i+1);
			regions.add(act_region);
		}
		
		// clean up
		rmap.clear();
		real_labels.clear();
		
		return new MTBRegion2DSet( regions, 0.0, 0.0, img.getSizeX()-1.0, img.getSizeY()-1.0);
	}
	
	/** returns the value of labelImg at position (x,y).
	 * Also checks for image boundaries, returns 0 when (x,y) is outside of the image
	 * @param x x-coordinate
	 * @param y y-coordinate
	 * @param labelImg actual label image
	 * @return pixel value at (x,y) if (x,y) inside the image. If outside of the image, returns 0
	 */
	private int getLabel(int x, int y) {
		
		if ((x >= 0) && (x < m_width) && (y >= 0) && (y < m_height)) {
			return m_labelImg.getValueInt(x,y);
		}
		else {
			return 0;
		}
	}
	
	
	
	
	/**
	 * @param binIP				binary input image
	 * @param diagonalNeighbors	set true for 8-neighborhood components, set false for 4-neighborhood
	 * @return 					returns a vector of regions	
	 */
	@Deprecated
	public Vector<MTBRegion2D> regionSegmentation(ImageProcessor binIP, boolean diagonalNeighbors) {
		
		m_width= binIP.getWidth();
		m_height= binIP.getHeight();
		
		Vector<MTBRegion2D> regions= new Vector<MTBRegion2D>();
		
		m_labelImage= 
				NewImage.createFloatImage("Label Image", m_width, m_height, 1, NewImage.FILL_BLACK);
        m_labelIP= m_labelImage.getProcessor();
		
        
        // western, northwestern, northern, northeastern label value of the actual pixel
        int l_w,l_nw,l_n,l_ne;
        
        // if only 4-neighborhood is regarded
        // set the diagonal neighors to zero to leave them
        // out of consideration in the following algorithm
        l_nw = 0;
        l_ne = 0;
        
        // smallest neigbor-label
        int l_min;
        
        // count of labeled neighbors
        int l_cnt;
        
        // label variable, which is incremented for new labels
		int label = 1;
		
		
		// collision sets for certain labels (vectorindex = label-1)
		Vector< LinkedHashSet<Integer> > collisions = new Vector< LinkedHashSet<Integer> >();
		
		// tmp references to collision sets
		LinkedHashSet<Integer> tmpSet1;
		LinkedHashSet<Integer> tmpSet2;
		// Iterator over the set elements
		Iterator<Integer> setIter;
		Integer labelInt;
		
		// first step: 
		// run line by line through the image
		// and label pixels		
		for (int y = 0; y < m_height; y++) {
			for (int x = 0; x < m_width; x++) {
				
				if (binIP.getPixel(x,y) > 0 && m_labelIP.getPixel(x,y)==0) {
					
					l_cnt = 0;
					l_min = 0;
					
					if (diagonalNeighbors) {
						
						// get relevant neighboring labels (8-neighborhood)
						l_w = getLabelSavely(x-1,y);
						l_nw = getLabelSavely(x-1,y-1);
						l_n = getLabelSavely(x,y-1);
						l_ne = getLabelSavely(x+1,y-1);
					
						// determine smallest label and count labeled neighbors
						if (l_w > 0) {
							l_cnt++;
							l_min = l_w;
						}
						
						if (l_nw > 0) {
							l_cnt++;
							
							if (l_min == 0 || l_nw < l_min) {
								l_min = l_nw;
							}
						}
						
						if (l_n > 0) {
							l_cnt++;
							
							if (l_min == 0 || l_n < l_min) {
								l_min = l_n;
							}
						}
						
						if (l_ne > 0) {
							l_cnt++;
						
							if (l_min == 0 || l_ne < l_min) {
								l_min = l_ne;
							}
						}						
					}
					else {
						
						// get relevant neighboring labels (4-neighborhood)
						l_w = getLabelSavely(x-1,y);
						l_n = getLabelSavely(x,y-1);
					
						// determine smallest label and count labeled neighbors
						if (l_w > 0) {
							l_cnt++;
							l_min = l_w;
						}
						
						if (l_n > 0) {
							l_cnt++;
							
							if (l_min == 0 || l_n < l_min) {
								l_min = l_n;
							}
						}									
					}

					if (l_cnt == 0) {
					// no neighbors of interest are labeled
						
						// give pixel a new label
						m_labelIP.putPixel(x,y,label);
						
						// create the collision set for that label
						// only containing the new label (-> no collision)
						tmpSet1 = new LinkedHashSet<Integer>();
						tmpSet1.add(label);
						
						// add the new collision set to the vector of collision sets
						collisions.add(tmpSet1);
						
						label++;
					}
					else if (l_cnt == 1) {
					// no label-collision, only one neighbor is labeled, adopt the label
						
						m_labelIP.putPixel(x, y, l_min);
					}
					else {
					// there might be a label-collision! More than one neighbor is labeled	
						
						
						// assign lowest label to the actual pixel
						m_labelIP.putPixel(x, y, l_min);
						
						// get the collision set of the label l_min
						tmpSet1 = collisions.get(l_min - 1);
						
						// test for label collisions
						if (l_w > 0 && l_w != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_w - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}
								
						if (l_nw > 0 && l_nw != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_nw - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}						
						
						if (l_n > 0 && l_n != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_n - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}					
																						
						if (l_ne > 0 && l_ne != l_min) {
							
							// get collision set of the colliding label
							tmpSet2 = collisions.get(l_ne - 1);
							
							// if the collision sets are not already the same, join them
							if (tmpSet2 != tmpSet1) {
								setIter = tmpSet2.iterator();
								
								while (setIter.hasNext()) {
									labelInt = setIter.next();
									tmpSet1.add(labelInt);
									collisions.set(labelInt.intValue() - 1, tmpSet1);
								}
							}	
						}											
					}
				}
			}
		}
		
		// run over collision sets and build a simple label mapping array (labelMap):
		// (index of the array - 1) corresponds to the found label
		// the array value corresponds to the value to which the found label is mapped
		int[] labelMapping = new int[label - 1];
		for (int i = 0; i < labelMapping.length; i++) {
			
			// get the collision set for label (i+1)
			tmpSet1 = collisions.get(i);
	
			// set the label mapping to the first label in the set
			setIter = tmpSet1.iterator();
			labelMapping[i] = setIter.next().intValue();		
		}
		
		// clean up
		collisions.clear();
		
		
		// second step: 
		// run line by line through the image and resolve the label collisions,
		// create Region objects and compute its properties
		
		// relates a label (int) to a Region object
		HashMap<Integer,MTBRegion2D> rmap = new HashMap<Integer,MTBRegion2D>();
		Vector<Integer> real_labels = new Vector<Integer>();
		MTBRegion2D act_region;
		int real_label;
		
		for (int y = 0; y  <m_height; y++) {
			for (int x = 0; x < m_width; x++) {
				
				label = m_labelIP.getPixel(x,y);
				
				// if a labeled pixel is found
				if (label != 0) {
					
					real_label = labelMapping[label - 1];
					
					if (rmap.containsKey(real_label)) {
					// update region					
						
						act_region = rmap.get(real_label);
						act_region.addPixel(x, y);						
					}
					else {
					// new region
						
						act_region = new MTBRegion2D();
						act_region.addPixel(x, y);	
						rmap.put(real_label, act_region);
						
						real_labels.add(real_label);
					}				
				}
			}
		}
			
		// build vector of regions
		for (int i = 0; i < real_labels.size(); i++) {
			regions.add(rmap.get(real_labels.get(i)));
		}
		
		// clean up
		rmap.clear();
		real_labels.clear();
		
		return regions;
	}
	
	
	/** returns the value of labelImg at position (x,y).
	 * Also checks for image boundaries, returns 0 when (x,y) is outside of the image
	 * @param x x-coordinate
	 * @param y y-coordinate
	 * @param labelImg actual label image
	 * @return pixel value at (x,y) if (x,y) inside the image. If outside of the image, returns 0
	 */
	@Deprecated
	private int getLabelSavely(int x, int y) {
		
		if ((x >= 0) && (x < m_width) && (y >= 0) && (y < m_height)) {
			return m_labelIP.getPixel(x,y);
		}
		else {
			return 0;
		}
	}
}
