/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.thresholds;

import java.util.LinkedList;

import de.unihalle.informatik.MiToBo.datatypes.*;
import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.operator.*;

/**
 * class for calculating global threshold using the 
 * moment preserving method [1]
 * 
 * [1] 	W. Tsai, Moment-preserving thresholding: A new approach, 
 * 		Comput. Vision Graphics Image Process. 29, 1985, 377-393.
 * 
 * @author glass
 *
 */
public class CalcGlobalThreshMoments extends MTBOperator
{
  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

	MTBOpArgumentDescriptor[] inputDescriptors = 
		new MTBOpArgumentDescriptor[] {
			new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
					"Input image", true, null)
	};	
	MTBOpArgumentDescriptor[] outputDescriptors = 
		new MTBOpArgumentDescriptor[] {
			new MTBOpArgumentDescriptor("resultThreshold", MTBDoubleData.class, 
					"Resulting threshold", true, null)
	};
	MTBOpArgumentDescriptor[] parameterDescriptors = null;
	MTBOpArgumentDescriptor[] supplementalDescriptors = null;

	
		addDescriptors( parameterDescriptors, inputDescriptors, 
				outputDescriptors, supplementalDescriptors);
	}

	public CalcGlobalThreshMoments() throws MTBOperatorException {
		// nothing to do here
	}

	public CalcGlobalThreshMoments(MTBImage img) throws MTBOperatorException {
		this.setInput("inImg", img);
	}

	/**
	 * Get reference to the current input image.
	 * 
	 * @return	Input image to work on.
	 * @throws MTBOperatorException 
	 */
	public MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)(this.getInput("inImg"));
	}
	
	/**
	 * Get the resulting threshold.
	 * 
	 * @return	Moments threshold.
	 * @throws MTBOperatorException 
	 */
	public double getThreshold() throws MTBOperatorException {
		return ((MTBDoubleData)(this.getOutput("resultThreshold"))).getValue().doubleValue();
	}
	
	
	/**
	 * Set result threshold.
	 * 
	 * @param thresh	Resulting threshold.
	 * @throws MTBOperatorException 
	 */
	public void setResultThresh(double thresh) throws MTBOperatorException {
		this.setOutput("resultThreshold", new MTBDoubleData(new Double(thresh)));
	}

	@Override
	protected void operate() throws MTBOperatorException {
		double threshold = this.calcThreshold(this.getInputImage());
		this.setResultThresh(threshold);
	}
	
	private int calcThreshold(MTBImage inputImg) {
		MTBHistogram hist = new MTBHistogram(inputImg);
		return calcThreshold(hist);
	}
	
	private int calcThreshold(MTBHistogram hist)
	{
		double l = hist.getSize();
		
		int thres = -1;
		
		double m0 = 1;
		double m1 = 0;
		double m2 = 0;
		double m3 = 0;
		
		double cd, c0, c1;
		double z0, z1;
		double p0 = 0;
		
		hist.normalize();
		double[] data = hist.getData();

		// calculate first, second and third order moments
		for(int i = 0; i < data.length; i++)
		{		
			double val = i/l;
			
			m1 += val * data[i];
			m2 += val * val * data[i];
			m3 += val * val * val * data[i];
		}
		
		cd = m0 * m2 - m1 * m1;
		c0 = ( -m2 * m2 + m1 * m3 ) / cd;
		c1 = ( m0 * -m3 + m2 * m1 ) / cd;
		z0 = 0.5 * ( -c1 - Math.sqrt ( c1 * c1 - 4.0 * c0 ) );
		z1 = 0.5 * ( -c1 + Math.sqrt ( c1 * c1 - 4.0 * c0 ) );
		p0 = ( z1 - m1 ) / ( z1 - z0 );  
		
		// The threshold is the gray-level closest  
		// to the p0-tile of the normalized histogram 
		double sum = 0;
		
		for (int i = 0; i < data.length; i++)
		{
			sum += data[i];
			
			if (sum > p0) 
			{
				thres = i;
				break;
			}
		}
		return thres;
	}
}
