/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.thresholds;

import java.util.LinkedList;

import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.operator.*;
import de.unihalle.informatik.MiToBo.datatypes.*;
import de.unihalle.informatik.MiToBo.datatypes.images.*;

/**
 * This class implements routines to calculate Otsu's
 * threshold from given grayscale images. 
 * <p>
 * Basically, a threshold is extracted that separates the image
 * histogram in two classes, more or less equal in size
 * and well-separated in terms of their average intensities.
 * <p>
 * Otsu thresholding is based on the following goal function,
 * 
 * {@latex.ilb %preamble{\\usepackage{amssymb}}
 * 		  \\begin{eqnarray*}
 *  		J(T) &=& \\pi_{fg} \\cdot \\pi_{bg} \\cdot ( \\mu_{fg} - \\mu_{bg} )^2, \\\\
 *        \\end{eqnarray*}}
 * 
 * which is to be maximized with regard to threshold T.
 * 
 * @author moeller
 */
public class CalcGlobalThreshOtsu extends MTBOperator {
	
  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("inImg", MTBImage.class, "Input image", true, null)
			};
	
			MTBOpArgumentDescriptor[] outputDescriptors = 
				new MTBOpArgumentDescriptor[] {
					new MTBOpArgumentDescriptor("Otsu_threshold", MTBDoubleData.class, "Result image", true, null)
			};
	
			
				addDescriptors( null, inputDescriptors, 
						outputDescriptors, null);
		}

		public CalcGlobalThreshOtsu() throws MTBOperatorException {
			// nothing to do here
		}

	public CalcGlobalThreshOtsu(MTBImage inImg) throws MTBOperatorException {
		super();
		this.setInput("inImg", inImg);
	}

	/**
	 * Get reference to the current input image.
	 * 
	 * @return	Input image to work on.
	 */
	public MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)(this.getInput("inImg"));
	}
	/**
	 * Get the resulting threshold.
	 * 
	 * @return	Otsu threshold.
	 */
	public MTBDoubleData getOtsuThreshold() throws MTBOperatorException {
		return ((MTBDoubleData)(this.getOutput("Otsu_threshold")));
	}
		
	/**
	 * Set result threshold.
	 * 
	 * @param thresh	Resulting threshold.
	 */
	public void setResultThresh(double thresh) throws MTBOperatorException {
		this.setOutput("Otsu_threshold", new MTBDoubleData(thresh));
	}

	@Override
	protected void operate() throws MTBOperatorException {
		double threshold= this.calcThreshold( getInputImage());
		setResultThresh(threshold);
	}

	/**
	 * Calculate Otsu threshold given a 1D-histogram.
	 * 
	 * @param	histo	MiToBo image histogram
	 * @return	Calculated Otsu threshold.
 	 */
	private double calcThreshold(MTBHistogram histo) {

		// normalize histogram
		histo.normalize();
		
		// calculate mean; function in Histogram errorneous...
		double mean= histo.getMean();
		
		// search threshold
		double j_max= -1;
		
		int minBinID= histo.getSmallestNonEmptyBin();
		int maxBinID= histo.getLargestNonEmptyBin();
		double [] histoData= histo.getData();
		
		double thresh= 0;
		for (int t=minBinID; t<=maxBinID; ++t) {
			
			double omega_1= 0;
			for (int i=minBinID; i<=t; ++i)
				omega_1= omega_1 + histoData[i];
			double omega_2= 0;
			for (int i=t+1; i<=maxBinID; ++i)
				omega_2= omega_2 + histoData[i];
				
			double mu_1= 0;
			for (int i=minBinID; i<=t; ++i)
				mu_1= mu_1 + i*histoData[i]/omega_1;
			double mu_2= 0;
			for (int i=t+1; i<=maxBinID; ++i)
				mu_2= mu_2 + i*histoData[i]/omega_2;
			
			if (omega_1*(mu_1-mean)*(mu_1-mean) + omega_2*(mu_2-mean)*(mu_2-mean) > j_max) {
				j_max= omega_1*(mu_1-mean)*(mu_1-mean) + omega_2*(mu_2-mean)*(mu_2-mean);
				thresh= t;
			}
		}
		return thresh;
	}
	
	private double calcThreshold(MTBImage mtbImage) {
		MTBHistogram imageHisto= new MTBHistogram(mtbImage);
		return this.calcThreshold(imageHisto);
	}
	
}
