/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_io;

import ij.IJ;
import ij.plugin.PlugIn;

import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Locale;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.WindowConstants;
import javax.swing.border.Border;

import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.io.files.ImageOpen;
import de.unihalle.informatik.MiToBo.io.files.ReaderMTBImage;
import de.unihalle.informatik.MiToBo.io.tools.ImageFilter;
import de.unihalle.informatik.MiToBo.tools.system.EnvironmentConfig;

/**
 * ImageJ plugin to open a single image, several images or images from a given
 * directory. From every image to open, all or several channels can be used to
 * open. The channels should be given in a form like 0,1,2,3-5 for example, to
 * open the channels 0,1,2,3,4,5 from all images to open. The given channels 
 * used for all images, it is not possible to open several images with several
 * channels. To realize this, images should be opened apart.
 * <p>
 * Multiple file/directory selection is allowed. If several files or 
 * directories selected, all selected images and all images in the selected 
 * directory will be opened. Directory search is not recursive.
 * <p>
 * The directory where browsing starts the first time after the start of 
 * ImageJ is determined by environment variable "MITOBO_OPEN_IMAGE_OPENDIR",
 * the java property "mitobo.open_image.opendir" or the corresponding ImageJ
 * property, the environment variable "MITOBO_IMAGEDIR", the java 
 * property "mitobo.open_image.imagedir" or the corresponding ImageJ property,
 * in this order. If none is set, the recently visited directory is used,
 * or if not specified so far, the home directory of the user.
 * 
 * @see ImageOpen
 * 
 * @modified 28.05.2010
 * 
 * @author misiak
 */

public class Open_Image extends JFrame implements PlugIn, ActionListener {
  /**
   * SerialVersionUID.
   */
  private static final long serialVersionUID = 1L;
  /**
   * File choose dialog to select images and directories to open.
   */
  private JFileChooser fc;
  /**
   * List of files to open.
   */
  private File[] inputFiles;

  /**
   * Channels to open from the image.
   */
  private int[] channels;
  /**
   * Get initial directory from environment configuration.
   */
  static String currentDir;
  static {
    String dir;
    dir = EnvironmentConfig.getConfigValue("open_image", "opendir");
    System.out.println(dir);
    if (dir == null) {
    	dir = EnvironmentConfig.getConfigValue(null, "imagedir");
      if (dir == null) {
      	dir= ij.Prefs.get("dir.image", null);
      	if (dir == null)
      		currentDir = System.getProperty("user.home");
      	else
      		currentDir = dir;
      }
      else 
      	currentDir = dir;
    }
    else
    	currentDir = dir;
  }

  /**
   * Plugin GUI window.
   */
  private JFrame window;
  /**
   * Defined color.
   */
  private Color myBlue = new Color(184, 207, 229);
  /**
   * Return value from file chooser dialog.
   */
  private int returnVal;
  /**
   * Text fields for file/directory input and the list of channels to open.
   */
  private JTextField inputField, chanField;
  /**
   * Several buttons.
   */
  private JButton browse, open, cancel;
  /**
   * Radio buttons to choose whether all channels should be opened or some
   * specific ones.
   */
  private JRadioButton openAll, openSome;

  /**
   * Main run method for plugin.
   */
  @Override
  public void run(String arg) {
    this.inputFiles = new File[0];
    this.channels = new int[0];
    this.returnVal = 0;
    controlWindow();
  }

  /**
   * Open method to open the selected files/directories.
   */
  public void open() {
//    IndexParser IP = new IndexParser();
//    if (this.chanField.getText().length() > 0) {
//      this.channels = IP.parseIndices(this.chanField.getText());
//    }
    for (int i = 0; i < this.inputFiles.length; i++) {
      if (this.inputFiles[i].isDirectory()) {
//        ImageDirOpen IDO = new ImageDirOpen();
//        if (this.channels.length > 0) {
//          IDO = new ImageDirOpen(this.inputFiles[i], this.channels);
//        } else {
//          IDO = new ImageDirOpen(this.inputFiles[i]);
//        }
//        MTBImage[] theImages = new MTBImage[0];
//        try {
//          theImages = IDO.open();
//        } catch (MTBProcessingDAGException e) {
//          // TODO Auto-generated catch block
//          e.printStackTrace();
//        }
//        if (theImages.length > 0) {
//          for (int j = 0; j < theImages.length; j++) {
//            theImages[j].show();
//          }
//        } else {
//          System.out.println("   --> open: no images to open in "
//              + this.inputFiles[i]);
//        }
    	  
    	  File[] files = this.inputFiles[i].listFiles(new ImageFilter(new String[]{"tif", "tiff", "TIFF", "TIF"}));
    	  
    	  ReaderMTBImage reader = null;
    	  
    	  for (int j = 0; j < files.length; j++) { 
    		if (files[j].isFile()) {  
    		  try {
				reader = new ReaderMTBImage(files[j].getAbsolutePath());
				reader.runOp(null);
				reader.getResultImage().show();
    		  } catch (MTBOperatorException e) {
				IJ.error("Cannot open " + files[j].getAbsolutePath() + ": " + e.getMessage());
				e.printStackTrace();
    		  } catch (MTBProcessingDAGException e) {
  				IJ.error("Cannot open " + files[j].getAbsolutePath() + ": " + e.getMessage());
				e.printStackTrace();
			}
    	  }
    	}
    	  
      } else if (this.inputFiles[i].isFile()) {
//        MTBImage theImage = null;
//        if (this.channels.length > 0) {
//          try {
//            ImageOpen IO = new ImageOpen(this.inputFiles[i].getPath(), this.channels,
//                true, 0);
//            IO.runOp(null);
//            theImage = IO.getImage();
//          } catch (MTBOperatorException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
//          } catch (MTBProcessingDAGException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
//          }
//        } else {
//          try {
//        	  
//        	  
//        	  
//            ImageOpen_new IO = new ImageOpen_new(this.inputFiles[i].getPath(), null, true, 0);
//            IO.runOp(null);
//            theImage = IO.getImage();
//          } catch (MTBOperatorException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
//          } catch (MTBProcessingDAGException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
//          }
//        }
//        if (theImage != null) {
//          theImage.show();
//        }
    	  
    	  ReaderMTBImage reader = null;
    	  try {
				reader = new ReaderMTBImage(this.inputFiles[i].getAbsolutePath());
				reader.runOp(null);
				reader.getResultImage().show();
  		  } catch (MTBOperatorException e) {
				IJ.error("Cannot open " + this.inputFiles[i].getAbsolutePath() + ": " + e.getMessage());
				e.printStackTrace();
  		  } catch (MTBProcessingDAGException e) {
				IJ.error("Cannot open " + this.inputFiles[i].getAbsolutePath() + ": " + e.getMessage());
				e.printStackTrace();
			}
    	  
      } else {
        System.out.println("   --> open: cannot open "
            + this.inputFiles[i].getAbsolutePath() + ", no such file or directory");
      }
    }
  }

  /**
   * GUI window for plugin to select files/directories and make some settings
   * for the opener.
   */
  public void controlWindow() {
    /*
     * main window
     */
    this.window = new JFrame("MiToBo - Open_Image | Control Window");
    this.window.setSize(440, 220);
    this.window.setLocationRelativeTo(null);
    this.window.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    Border setBorder = BorderFactory.createLineBorder(this.myBlue, 2);
    /*
     * main panel
     */
    JPanel guiPanel = new JPanel();
    guiPanel.setLayout(new FlowLayout(0, 10, 10));
    /*
     * panel1 - input panel
     */
    JPanel panel1 = new JPanel();
    panel1.setBorder(BorderFactory
        .createTitledBorder(setBorder, "Browse Input"));
    panel1.setLayout(new GridLayout(1, 1));
    JPanel panel11 = new JPanel();
    panel11.setLayout(new FlowLayout(0, 10, 10));
    this.inputField = new JTextField(25);
    this.inputField.addActionListener(this);
    this.inputField.setText(currentDir);
    this.inputField.setToolTipText("Path to input image or directory.");
    this.browse = new JButton("Browse");
    this.browse.setActionCommand("browse");
    this.browse.addActionListener(this);
    panel11.add(this.inputField);
    panel11.add(new JLabel("  "));
    panel11.add(this.browse);
    panel1.add(panel11);
    guiPanel.add(panel1);
    /*
     * panel2 - settings panel
     */
//    JPanel panel2 = new JPanel();
//    panel2.setBorder(BorderFactory.createTitledBorder(setBorder,
//        "Image Settings"));
//    panel2.setLayout(new FlowLayout(0, 10, 10));
//    JPanel panel22 = new JPanel();
//    panel22.setLayout(new GridLayout(2, 2));
//    this.openAll = new JRadioButton("Open All Channels");
//    this.openAll.setActionCommand("all");
//    this.openAll.addActionListener(this);
//    this.openSome = new JRadioButton("Open Several Channels");
//    this.openSome.setActionCommand("some");
//    this.openSome.addActionListener(this);
//    ButtonGroup bg = new ButtonGroup();
//    bg.add(this.openAll);
//    bg.add(this.openSome);
//    bg.setSelected(this.openAll.getModel(), true);
//    this.chanField = new JTextField(17);
//    this.chanField.setText("");
//    this.chanField
//        .setToolTipText("Channels to open from all images.\n Insert like 0,1,3,5-12.");
//    this.chanField.setEditable(false);
//    panel22.add(this.openAll);
//    panel22.add(new JLabel("   "));
//    panel22.add(this.openSome);
//    panel22.add(this.chanField);
//    panel2.add(panel22);
//    guiPanel.add(panel2);
    /*
     * panel3 - open and cancel button
     */
    JPanel panel3 = new JPanel();
    panel3.setBorder(BorderFactory.createTitledBorder(setBorder, "Run Opener"));
    panel3.setLayout(new GridLayout(1, 2));
    JPanel panel33 = new JPanel();
    panel3.setLayout(new FlowLayout(0, 126, 10));
    this.open = new JButton("Open");
    this.open.setActionCommand("open");
    this.open.addActionListener(this);
    this.cancel = new JButton("Cancel");
    this.cancel.setActionCommand("cancel");
    this.cancel.addActionListener(this);
    panel33.add(this.open);
    panel33.add(this.cancel);
    panel3.add(panel33);
    guiPanel.add(panel3);
    /*
     * add to window
     */
    this.window.add(guiPanel);
    this.window.setResizable(false);
    this.window.setVisible(true);
  }

  /**
   * Handle action events from fields and buttons.
   */
  @Override
  public void actionPerformed(ActionEvent e) {
    /*
     * Browse button was pressed.
     */
    if (e.getActionCommand() == "browse") {
      // set the language from the file choose to US-English
      Locale.setDefault(Locale.US);
      JComponent.setDefaultLocale(Locale.US);
      this.fc = new JFileChooser(Open_Image.currentDir);
      this.fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
      // multiple files can be selected, images as well as directories
      this.fc.setMultiSelectionEnabled(true);
      // adding file filter for image files (see io.files.ImageFiler class)
      this.fc.addChoosableFileFilter(new ImageFilter(new String[]{"tif", "tiff", "TIF", "TIFF"}));
      this.fc.setAcceptAllFileFilterUsed(false);
      this.fc.setVisible(true);
      this.returnVal = this.fc.showOpenDialog(this.fc);
      if (this.returnVal == JFileChooser.APPROVE_OPTION) {
        this.inputFiles = this.fc.getSelectedFiles();
        currentDir = this.inputFiles[0].getParent();
        // remember currentDir in ImageJ preferences
        EnvironmentConfig.setImageJPref("open_image", "opendir", currentDir);
        if (this.inputFiles.length > 1) {
          String text = "";
          for (int i = 0; i < this.inputFiles.length - 1; i++) {
            text += "\"" + this.inputFiles[i].getName() + "\" ";
          }
          text += "\"" + this.inputFiles[this.inputFiles.length - 1].getName() + "\"";
          this.inputField.setText(text);
        } else {
          this.inputField.setText(this.inputFiles[0].getAbsolutePath());
        }
      }
    }
    /**
     * Choose all channels from the images to open.
     */
    if (e.getActionCommand() == "all") {
      this.chanField.setEditable(false);
    }
    /*
     * Choose several channels to open from all image. Channels are given from
     * the text field.
     */
    if (e.getActionCommand() == "some") {
      this.chanField.setEditable(true);
    }
    /*
     * Opener was canceled by user.
     */
    if (e.getActionCommand() == "cancel") {
      this.window.dispose();
      System.out.println("   --> Open_image canceled.");
    }
    /*
     * Open selected image(s) or directory.
     */
    if (e.getActionCommand() == "open") {
      if (this.inputFiles.length == 0) {
        File theFile = new File(this.inputField.getText());
        if (theFile.exists()) {
          this.inputFiles = new File[1];
          this.inputFiles[0] = theFile;
          if (dirWarning()) {
            open();
            this.window.dispose();
          }
        } else {
          JOptionPane.showMessageDialog(this, "Cannot open "
              + theFile.getAbsolutePath() + ", no such file or directory",
              "Error Open File", JOptionPane.ERROR_MESSAGE);
        }
      } else {
        if (dirWarning()) {
          open();
          this.window.dispose();
        }
      }
    }
  }

  public boolean dirWarning() {
    boolean warning = false;
    int value = 0;
    Locale.setDefault(Locale.US);
    JComponent.setDefaultLocale(Locale.US);
    for (int i = 0; i < this.inputFiles.length; i++) {
      if (this.inputFiles[i].isDirectory()) {
        value = JOptionPane
            .showConfirmDialog(
                this,
                "A directory has been selected.\nDo you really want to open ALL included image files ?",
                "Warning Open Directory", JOptionPane.YES_NO_OPTION,
                JOptionPane.WARNING_MESSAGE);
        break;
      }
    }
    if (value == JOptionPane.NO_OPTION) {
      warning = false;
    } else if (value == JOptionPane.YES_OPTION) {
      warning = true;
    }
    return warning;
  }
}
