/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_io;

import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;
import java.io.File;
import java.util.Locale;
import javax.swing.*;

import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.io.files.ImageSave;
import de.unihalle.informatik.MiToBo.io.files.WriterMTBImage;
import de.unihalle.informatik.MiToBo.io.tools.ImageFilter;
import de.unihalle.informatik.MiToBo.tools.system.EnvironmentConfig;

/**
 * ImageJ plugin to save images from type tif/tiff;jpg/jpeg;png;gif;lif;ppm;
 * pgm. Only tif files can be saved as stacks.
 * <p>
 * The directory where browsing starts the first time after the start of 
 * ImageJ is determined by environment variable "MITOBO_SAVE_IMAGE_SAVEDIR",
 * the java property "mitobo.save_image.savedir" or the corresponding ImageJ
 * property, the environment variable "MITOBO_IMAGEDIR", the java 
 * property "mitobo.open_image.imagedir" or the corresponding ImageJ property,
 * in this order. If none is set, the recently visited directory is used,
 * or if not specified so far, the home directory of the user.
 * 
 * @see ImageSave
 * 
 * @modified 10.05.2010
 * 
 * @author misiak
 */

public class Save_Image extends JFrame implements PlugInFilter {
  /**
   * SerialVersionUID.
   */
  private static final long serialVersionUID = 1L;
  /**
   * The file where the current image should be saved.
   */
  private File inputFile;
  /**
   * The MTBImage of the current opened ImagePlus image that should be saved.
   */
  private MTBImage inputImg;
  /**
   * The opened ImagePlus image to save.
   */
  private ImagePlus img;
  /**
   * Remember directory which was last opened.
   */
  private static String currentDir;
  static {
    String dir;

    dir = EnvironmentConfig.getConfigValue("save_image", "savedir");
    if (dir == null) {
    	// note: open_image as plugin name is correct, the variable is 
    	//       shared among both plugins!
    	dir = EnvironmentConfig.getConfigValue(null, "imagedir");
      if (dir == null) {
      	dir= ij.Prefs.get("dir.image", null);
      	if (dir == null)
      		currentDir = System.getProperty("user.home");
      	else
      		currentDir = dir;
      }
      else 
      	currentDir = dir;
    }
    else
    	currentDir = dir;
  }

  @Override
  public int setup(String arg, ImagePlus image) {
    this.img = image;
    return DOES_ALL;
  }

  @Override
  public void run(ImageProcessor ip) {
    save();
  }

  /**
   * Save method to save the current image. The method tests whether the file
   * exists and has a valid image-file extension.
   */
  public void save() {
    // create the MTBImage from the current ImagePlus image
    this.inputImg = MTBImage.createMTBImage(this.img);
    // create an file chooser dialoge
    final JFileChooser fc = new JFileChooser(currentDir);
    // add image file type filter
    fc.addChoosableFileFilter(new ImageFilter());
    fc.setAcceptAllFileFilterUsed(false);
    int returnVal = 0;
    boolean overwrite = false;
    boolean error = true;
    // test for errors (existence ore invalid extension)
    while (error) {
      // get the file where the image should be saved
      returnVal = fc.showSaveDialog(fc);
      // save dialog throws no errors and is not canceled
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        this.inputFile = fc.getSelectedFile();
        // test for existing files
        if (this.inputFile.exists()) {
          // warning message for overwriting files
          String[] options = { "Overwrite", "No" };
          Locale.setDefault(Locale.US);
          JComponent.setDefaultLocale(Locale.US);
          int warning = JOptionPane.showOptionDialog(this,
              "The file exists. Do you want to overwrite the file?",
              "File Exists!", JOptionPane.YES_NO_OPTION,
              JOptionPane.WARNING_MESSAGE, null, options, options[1]);
          // do not overwrite the existing file
          if (warning == JOptionPane.NO_OPTION) {
            error = true;
            overwrite = false;
          }
          // overwrite the existing file
          if (warning == JOptionPane.YES_OPTION) {
            ImageFilter IF = new ImageFilter(new String[]{"tif", "tiff", "TIFF", "TIF"});
            // test for valid image-file extension
            if (IF.accept(this.inputFile)) {
              error = false;
              overwrite = true;
            } else {
              JOptionPane.showMessageDialog(this,
                  "Invalid filename or image-file extension!", "Error",
                  JOptionPane.ERROR_MESSAGE);
              error = true;
              overwrite = false;
            }
          }
        } else {
          // file does not exist and test for valid image-file extension
          ImageFilter IF = new ImageFilter(new String[]{"tif", "tiff", "TIFF", "TIF"});
          if (IF.accept(this.inputFile)) {
            // file does not exist and has a valid extension, so save it
            error = false;
            overwrite = false;
            currentDir = fc.getCurrentDirectory().toString();
            // remember currentDir in ImageJ preferences
            EnvironmentConfig.setImageJPref("save_image", "savedir", currentDir);
            try {
            	WriterMTBImage writer = new WriterMTBImage(this.inputImg, this.inputFile.getAbsolutePath());
            	writer.setOverwrite(overwrite);
            	writer.runOp(null);
            	
           //   ImageSave IS = new ImageSave(this.inputFile.getPath(), this.inputImg, true, overwrite);
           //   IS.runOp(null);
            } catch (MTBOperatorException e) {
              e.printStackTrace();
            } catch (MTBProcessingDAGException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
          } else {
            /*
             * file does not exist but has a invalid file extension, so do not
             * save it
             */
            JOptionPane.showMessageDialog(this,
                "Invalid filename or image-file extension!", "Error",
                JOptionPane.ERROR_MESSAGE);
            error = true;
            overwrite = false;
          }
        }
      } else {
        // save dialog throws errors or is canceled
        error = false;
        overwrite = false;
        System.out.println("   --> " + "Error saving file: " + this.inputFile);
        System.out.println("   --> "
            + "------- image and meta-data NOT saved -------\n");
      }
    }
    // the file has a valid extension and an existing file should be replaced
    if (overwrite) {
      currentDir = fc.getCurrentDirectory().toString();
      // remember currentDir in ImageJ preferences
      EnvironmentConfig.setImageJPref("save_image", "savedir", currentDir);
      try {
        ImageSave IS = new ImageSave(this.inputFile.getPath(), this.inputImg, true, overwrite);
        IS.runOp(null);
      } catch (MTBOperatorException e) {
        e.printStackTrace();
      } catch (MTBProcessingDAGException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

    }
  }
}
