/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_regions;

import java.awt.*;
import java.awt.geom.Point2D;
import java.util.*;

import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.plugin.filter.*;
import ij.plugin.frame.*;

import de.unihalle.informatik.MiToBo.datatypes.*;
import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents.ContourType;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.*;

/**
 * 
 * @author moeller
 *
 * This class implements a 2D contour extraction on binary images.
 * <p>
 * Foreground pixels are assumed to have a value > 0,
 * all pixels with value 0 are thought as background.
 * The class calculates a label image and in addition extracts 
 * contours which are added to the ImageJ ROI manager.
 * Parts of the GUI have been copied from LabelComponents-Plugins.
 */
public class Extract_Contours implements PlugInFilter {
	
	/** original image */
	private MTBImage m_img;
	
	
	@Override
  public int setup(String arg, ImagePlus img) {
        try {
        	this.m_img = MTBImage.createMTBImage(img);
        } catch(IllegalArgumentException e) {
        	// do nothing because the below returned types are all legal.
        	// ImageJ will forbid an illegal type
        }       

		return DOES_8G+DOES_16+DOES_32;
	}
	
	/**
	 *  Main plugin method.
	 */
	@Override
  public void run(ImageProcessor ip) {

		try { 
			LabelComponentsSequential lcs = parameterDialog();
			lcs.setInput( "InputImage", this.m_img);
			lcs.runOp(null);
			MTBImage labelImg= lcs.getLabelImage();
			MTBImageByte byteLabelImg= 
				(MTBImageByte)labelImg.convertType(MTBImageType.MTB_BYTE, true);
			MTBRegion2DSet regions= lcs.getResultingRegions();
			
			ContourOnLabeledComponents contExtractor=
				new ContourOnLabeledComponents(byteLabelImg, regions, 
						ContourType.OUT_IN_CONTOUR, 50);
			contExtractor.runOp(null);
			MTBContour2DSet contours= contExtractor.getContours();
			
			System.out.println("Found " + contours.size() + " contours.");
			MTBImage resultImg= labelImg;
			RoiManager roiManage= RoiManager.getInstance();
			if (roiManage == null)
				roiManage= new RoiManager();
			for (int i=0;i<contours.size();++i) {
				// plot the contour
				resultImg= contours.elementAt(i).toMTBImage(null, resultImg);
				// add it to the ImageJ Roi Manager
				Vector<Point2D.Double> ppoints= contours.elementAt(i).getPoints();
				int [] xpoints= new int[ppoints.size()];
				int [] ypoints= new int[ppoints.size()];
				int j=0;
				for (Point2D.Double p: ppoints) {
					xpoints[j]= (int)p.x;
					ypoints[j]= (int)p.y;
					j++;
				}
				Polygon p= new Polygon(xpoints,ypoints,ppoints.size());
				Roi roi= new PolygonRoi(p,Roi.POLYGON);
				roiManage.add(resultImg.getImagePlus(),roi,i);
			}
			
			// rescale image
			int maxVal= (int)resultImg.getTypeMax();
			int maxVal_2= maxVal/2;
			for (int y=0;y<resultImg.getSizeY();++y) {
				for (int x=0;x<resultImg.getSizeX();++x) {
					if (   resultImg.getValueInt(x, y) < maxVal
							&& resultImg.getValueInt(x, y) > 0) {
						resultImg.putValueInt(x, y, maxVal_2);
					}
				}
			}
			resultImg.show();
		} catch (Exception e) {
			e.printStackTrace();
			IJ.showStatus("MiToBo Extract_Contours failed");
		}
	}
	
	// ---- Plugin helper functions ----
	
	/** Dialog window for parameter input 
	 * @return new LabelComponentsSequential3DParams object
	 */
	protected LabelComponentsSequential parameterDialog() 
		throws MTBOperatorException {
		
		// Dialog for parameter choice
		GenericDialog gd = new GenericDialog("Component labeling parameters");
		
		String[] neighOptions = {"4-Neighborhood", "8-Neighborhood"};
		gd.addChoice("Choose Neighborhood:", neighOptions, neighOptions[1]);	
		
		gd.showDialog();
		
		
		if (gd.wasOKed()) {
			
			// get neighborhood
			boolean diagonalNeighbors = 
				(gd.getNextChoiceIndex() == 0) ? false : true;	
			
			LabelComponentsSequential lcs = new LabelComponentsSequential();
			lcs.setParameter( "DiagonalNeighbors", new Boolean(diagonalNeighbors));
			return lcs;
			
		}
		return null;
	}
}
